from __future__ import annotations

from typing import TYPE_CHECKING, Any

from public import public

import ibis.expr.operations as ops
from ibis.common.deferred import deferrable
from ibis.expr.types.generic import Column, Scalar, Value

if TYPE_CHECKING:
    from collections.abc import Iterable, Mapping

    import ibis.expr.types as ir
    from ibis.expr.types.arrays import ArrayColumn


@public
class MapValue(Value):
    """A map literal or column expression.

    Can be constructed with [`ibis.map()`](#ibis.expr.types.map).

    Examples
    --------
    >>> import ibis
    >>> ibis.options.interactive = True
    >>> import pyarrow as pa
    >>> tab = pa.table(
    ...     {
    ...         "m": pa.array(
    ...             [[("a", 1), ("b", 2)], [("a", 1)], None],
    ...             type=pa.map_(pa.utf8(), pa.int64()),
    ...         )
    ...     }
    ... )
    >>> t = ibis.memtable(tab)
    >>> t
    ┏━━━━━━━━━━━━━━━━━━━━━━┓
    ┃ m                    ┃
    ┡━━━━━━━━━━━━━━━━━━━━━━┩
    │ map<!string, int64>  │
    ├──────────────────────┤
    │ {'a': 1, 'b': 2}     │
    │ {'a': 1}             │
    │ NULL                 │
    └──────────────────────┘

    Can use `[]` to access values:
    >>> t.m["a"]
    ┏━━━━━━━━━━━━━━━━━━━━━━┓
    ┃ MapGet(m, 'a', None) ┃
    ┡━━━━━━━━━━━━━━━━━━━━━━┩
    │ int64                │
    ├──────────────────────┤
    │                    1 │
    │                    1 │
    │                 NULL │
    └──────────────────────┘

    To provide default values, use `get`:
    >>> t.m.get("b", 0)
    ┏━━━━━━━━━━━━━━━━━━━┓
    ┃ MapGet(m, 'b', 0) ┃
    ┡━━━━━━━━━━━━━━━━━━━┩
    │ int64             │
    ├───────────────────┤
    │                 2 │
    │                 0 │
    │                 0 │
    └───────────────────┘
    """

    def get(
        self,
        key: ir.Value,
        default: ir.Value | None = None,
    ) -> ir.Value:
        """Return the value for `key` from `expr`.

        Return `default` if `key` is not in the map.

        Parameters
        ----------
        key
            Expression to use for key
        default
            Expression to return if `key` is not a key in `expr`

        Returns
        -------
        Value
            The element type of `self`

        Examples
        --------
        >>> import ibis
        >>> import pyarrow as pa
        >>> ibis.options.interactive = True
        >>> tab = pa.table(
        ...     {
        ...         "m": pa.array(
        ...             [[("a", 1), ("b", 2)], [("a", 1)], None],
        ...             type=pa.map_(pa.utf8(), pa.int64()),
        ...         )
        ...     }
        ... )
        >>> t = ibis.memtable(tab)
        >>> t
        ┏━━━━━━━━━━━━━━━━━━━━━━┓
        ┃ m                    ┃
        ┡━━━━━━━━━━━━━━━━━━━━━━┩
        │ map<!string, int64>  │
        ├──────────────────────┤
        │ {'a': 1, 'b': 2}     │
        │ {'a': 1}             │
        │ NULL                 │
        └──────────────────────┘
        >>> t.m.get("a")
        ┏━━━━━━━━━━━━━━━━━━━━━━┓
        ┃ MapGet(m, 'a', None) ┃
        ┡━━━━━━━━━━━━━━━━━━━━━━┩
        │ int64                │
        ├──────────────────────┤
        │                    1 │
        │                    1 │
        │                 NULL │
        └──────────────────────┘
        >>> t.m.get("b")
        ┏━━━━━━━━━━━━━━━━━━━━━━┓
        ┃ MapGet(m, 'b', None) ┃
        ┡━━━━━━━━━━━━━━━━━━━━━━┩
        │ int64                │
        ├──────────────────────┤
        │                    2 │
        │                 NULL │
        │                 NULL │
        └──────────────────────┘
        >>> t.m.get("b", 0)
        ┏━━━━━━━━━━━━━━━━━━━┓
        ┃ MapGet(m, 'b', 0) ┃
        ┡━━━━━━━━━━━━━━━━━━━┩
        │ int64             │
        ├───────────────────┤
        │                 2 │
        │                 0 │
        │                 0 │
        └───────────────────┘
        """

        return ops.MapGet(self, key, default).to_expr()

    def length(self) -> ir.IntegerValue:
        """Return the number of key-value pairs in the map.

        Returns
        -------
        IntegerValue
            The number of elements in `self`

        Examples
        --------
        >>> import ibis
        >>> import pyarrow as pa
        >>> ibis.options.interactive = True
        >>> tab = pa.table(
        ...     {
        ...         "m": pa.array(
        ...             [[("a", 1), ("b", 2)], [("a", 1)], None],
        ...             type=pa.map_(pa.utf8(), pa.int64()),
        ...         )
        ...     }
        ... )
        >>> t = ibis.memtable(tab)
        >>> t
        ┏━━━━━━━━━━━━━━━━━━━━━━┓
        ┃ m                    ┃
        ┡━━━━━━━━━━━━━━━━━━━━━━┩
        │ map<!string, int64>  │
        ├──────────────────────┤
        │ {'a': 1, 'b': 2}     │
        │ {'a': 1}             │
        │ NULL                 │
        └──────────────────────┘
        >>> t.m.length()
        ┏━━━━━━━━━━━━━━┓
        ┃ MapLength(m) ┃
        ┡━━━━━━━━━━━━━━┩
        │ int64        │
        ├──────────────┤
        │            2 │
        │            1 │
        │         NULL │
        └──────────────┘
        """
        return ops.MapLength(self).to_expr()

    def __getitem__(self, key: ir.Value) -> ir.Value:
        """Get the value for a given map `key`.

        ::: {.callout-note}
        ## This operation may have different semantics depending on the backend.

        Some backends return `NULL` when a key is missing, others may fail
        the query.
        :::

        Parameters
        ----------
        key
            A map key

        Returns
        -------
        Value
            An element with the value type of the map

        Examples
        --------
        >>> import ibis
        >>> import pyarrow as pa
        >>> ibis.options.interactive = True
        >>> tab = pa.table(
        ...     {
        ...         "m": pa.array(
        ...             [[("a", 1), ("b", 2)], [("a", 1)], None],
        ...             type=pa.map_(pa.utf8(), pa.int64()),
        ...         )
        ...     }
        ... )
        >>> t = ibis.memtable(tab)
        >>> t
        ┏━━━━━━━━━━━━━━━━━━━━━━┓
        ┃ m                    ┃
        ┡━━━━━━━━━━━━━━━━━━━━━━┩
        │ map<!string, int64>  │
        ├──────────────────────┤
        │ {'a': 1, 'b': 2}     │
        │ {'a': 1}             │
        │ NULL                 │
        └──────────────────────┘
        >>> t.m["a"]
        ┏━━━━━━━━━━━━━━━━━━━━━━┓
        ┃ MapGet(m, 'a', None) ┃
        ┡━━━━━━━━━━━━━━━━━━━━━━┩
        │ int64                │
        ├──────────────────────┤
        │                    1 │
        │                    1 │
        │                 NULL │
        └──────────────────────┘
        """
        return ops.MapGet(self, key).to_expr()

    def contains(
        self, key: int | str | ir.IntegerValue | ir.StringValue
    ) -> ir.BooleanValue:
        """Return whether the map contains `key`.

        Parameters
        ----------
        key
            Mapping key for which to check

        Returns
        -------
        BooleanValue
            Boolean indicating the presence of `key` in the map expression

        Examples
        --------
        >>> import ibis
        >>> import pyarrow as pa
        >>> ibis.options.interactive = True
        >>> tab = pa.table(
        ...     {
        ...         "m": pa.array(
        ...             [[("a", 1), ("b", 2)], [("a", 1)], None],
        ...             type=pa.map_(pa.utf8(), pa.int64()),
        ...         )
        ...     }
        ... )
        >>> t = ibis.memtable(tab)
        >>> t
        ┏━━━━━━━━━━━━━━━━━━━━━━┓
        ┃ m                    ┃
        ┡━━━━━━━━━━━━━━━━━━━━━━┩
        │ map<!string, int64>   │
        ├──────────────────────┤
        │ {'a': 1, 'b': 2}     │
        │ {'a': 1}             │
        │ NULL                 │
        └──────────────────────┘
        >>> t.m.contains("b")
        ┏━━━━━━━━━━━━━━━━━━━━━┓
        ┃ MapContains(m, 'b') ┃
        ┡━━━━━━━━━━━━━━━━━━━━━┩
        │ boolean             │
        ├─────────────────────┤
        │ True                │
        │ False               │
        │ False               │
        └─────────────────────┘
        """
        return ops.MapContains(self, key).to_expr()

    def keys(self) -> ir.ArrayValue:
        """Extract the keys of a map.

        Returns
        -------
        ArrayValue
            The keys of `self`

        Examples
        --------
        >>> import ibis
        >>> import pyarrow as pa
        >>> ibis.options.interactive = True
        >>> tab = pa.table(
        ...     {
        ...         "m": pa.array(
        ...             [[("a", 1), ("b", 2)], [("a", 1)], None],
        ...             type=pa.map_(pa.utf8(), pa.int64()),
        ...         )
        ...     }
        ... )
        >>> t = ibis.memtable(tab)
        >>> t
        ┏━━━━━━━━━━━━━━━━━━━━━━┓
        ┃ m                    ┃
        ┡━━━━━━━━━━━━━━━━━━━━━━┩
        │ map<!string, int64>  │
        ├──────────────────────┤
        │ {'a': 1, 'b': 2}     │
        │ {'a': 1}             │
        │ NULL                 │
        └──────────────────────┘
        >>> t.m.keys()
        ┏━━━━━━━━━━━━━━━━━━━━━━┓
        ┃ MapKeys(m)           ┃
        ┡━━━━━━━━━━━━━━━━━━━━━━┩
        │ array<!string>       │
        ├──────────────────────┤
        │ ['a', 'b']           │
        │ ['a']                │
        │ NULL                 │
        └──────────────────────┘
        """
        return ops.MapKeys(self).to_expr()

    def values(self) -> ir.ArrayValue:
        """Extract the values of a map.

        Returns
        -------
        ArrayValue
            The values of `self`

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> m = ibis.map({"a": 1, "b": 2})
        >>> m.values()
        [1, 2]
        """
        return ops.MapValues(self).to_expr()

    def __add__(self, other: MapValue) -> MapValue:
        """Concatenate this map with another.

        Parameters
        ----------
        other
            Map to concatenate with `self`

        Returns
        -------
        MapValue
            `self` concatenated with `other`

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> m1 = ibis.map({"a": 1, "b": 2})
        >>> m2 = ibis.map({"c": 3, "d": 4})
        >>> m1 + m2
        {'a': 1, 'b': 2, ... +2}
        """
        return ops.MapMerge(self, other).to_expr()

    def __radd__(self, other: MapValue) -> MapValue:
        """Concatenate this map with another.

        Parameters
        ----------
        other
            Map to concatenate with `self`

        Returns
        -------
        MapValue
            `self` concatenated with `other`

        Examples
        --------
        >>> import ibis
        >>> ibis.options.interactive = True
        >>> m1 = ibis.map({"a": 1, "b": 2})
        >>> m2 = ibis.map({"c": 3, "d": 4})
        >>> m1 + m2
        {'a': 1, 'b': 2, ... +2}
        """
        return ops.MapMerge(self, other).to_expr()


@public
class MapScalar(Scalar, MapValue):
    pass


@public
class MapColumn(Column, MapValue):
    def __getitem__(self, key: ir.Value) -> ir.Column:
        return MapValue.__getitem__(self, key)


@public
@deferrable
def map(
    keys: Iterable[Any] | Mapping[Any, Any] | ArrayColumn,
    values: Iterable[Any] | ArrayColumn | None = None,
) -> MapValue:
    """Create a [map container object](https://docs.python.org/3/glossary.html#term-mapping).

    If the `keys` and `values` are Python literals, then the output will be a
    `MapScalar`. If the `keys` and `values` are expressions (`ArrayColumn`),
    then the the output will be a `MapColumn`.

    Parameters
    ----------
    keys
        Keys of the map or `Mapping`. If `keys` is a `Mapping`, `values` must be `None`.
    values
        Values of the map or `None`. If `None`, the `keys` argument must be a `Mapping`.

    Returns
    -------
    MapValue
        An expression representing either a map column or literal (associative
        array with key/value pairs of fixed types)

    Examples
    --------
    Create a map literal from a dict with the type inferred

    >>> import ibis
    >>> ibis.options.interactive = True
    >>> ibis.map(dict(a=1, b=2))
    {'a': 1, 'b': 2}

    Create a new map column from columns with keys and values

    >>> import ibis
    >>> ibis.options.interactive = True
    >>> t = ibis.memtable({"keys": [["a", "b"], ["b"]], "values": [[1, 2], [3]]})
    >>> t
    ┏━━━━━━━━━━━━━━━━━━━━━━┳━━━━━━━━━━━━━━━━━━━━━━┓
    ┃ keys                 ┃ values               ┃
    ┡━━━━━━━━━━━━━━━━━━━━━━╇━━━━━━━━━━━━━━━━━━━━━━┩
    │ array<string>        │ array<int64>         │
    ├──────────────────────┼──────────────────────┤
    │ ['a', 'b']           │ [1, 2]               │
    │ ['b']                │ [3]                  │
    └──────────────────────┴──────────────────────┘
    >>> ibis.map(t.keys, t.values)
    ┏━━━━━━━━━━━━━━━━━━━━━━┓
    ┃ Map(keys, values)    ┃
    ┡━━━━━━━━━━━━━━━━━━━━━━┩
    │ map<string, int64>   │
    ├──────────────────────┤
    │ {'a': 1, 'b': 2}     │
    │ {'b': 3}             │
    └──────────────────────┘
    """
    if values is None:
        keys, values = tuple(keys.keys()), tuple(keys.values())
    return ops.Map(keys, values).to_expr()



from __future__ import annotations

import pandas as pd
import pandas.testing as tm

import ibis


def test_memtable_join(con):
    t1 = ibis.memtable({"x": [1, 2, 3], "y": [4, 5, 6], "z": ["a", "b", "c"]})
    t2 = ibis.memtable({"x": [3, 2, 1], "y": [7, 8, 9], "z": ["d", "e", "f"]})
    expr = t1.join(t2, "x")

    result = con.execute(expr)
    expected = pd.DataFrame(
        {
            "x": [1, 2, 3],
            "y": [4, 5, 6],
            "z": ["a", "b", "c"],
            "y_right": [9, 8, 7],
            "z_right": ["f", "e", "d"],
        }
    )

    left = result.sort_values("x").reset_index(drop=True)
    right = expected.sort_values("x").reset_index(drop=True)
    tm.assert_frame_equal(left, right)



from __future__ import annotations

import contextlib
import operator
from functools import singledispatch

import sqlglot as sg
import sqlglot.expressions as sge
import sqlglot.optimizer as sgo
import sqlglot.planner as sgp
from public import public

import ibis
import ibis.common.exceptions as com
import ibis.expr.datatypes as dt
import ibis.expr.schema as sch
import ibis.expr.types as ir
from ibis.util import experimental


class Catalog(dict[str, sch.Schema]):
    """A catalog of tables and their schemas."""

    typemap = {
        dt.Int8: "tinyint",
        dt.Int16: "smallint",
        dt.Int32: "int",
        dt.Int64: "bigint",
        dt.Float16: "halffloat",
        dt.Float32: "float",
        dt.Float64: "double",
        dt.Decimal: "decimal",
        dt.Boolean: "boolean",
        dt.JSON: "json",
        dt.Interval: "interval",
        dt.Timestamp: "datetime",
        dt.Date: "date",
        dt.Binary: "varbinary",
        dt.String: "varchar",
        dt.Array: "array",
        dt.Map: "map",
        dt.UUID: "uuid",
        dt.Struct: "struct",
    }

    def to_sqlglot_dtype(self, dtype: dt.DataType) -> str:
        if dtype.is_geospatial():
            return dtype.geotype
        else:
            default = dtype.__class__.__name__.lower()
            return self.typemap.get(type(dtype), default)

    def to_sqlglot_schema(self, schema: sch.Schema) -> dict[str, str]:
        return {name: self.to_sqlglot_dtype(dtype) for name, dtype in schema.items()}

    def to_sqlglot(self):
        return {
            name: self.to_sqlglot_schema(table.schema()) for name, table in self.items()
        }

    def overlay(self, step):
        updates = {dep.name: convert(dep, catalog=self) for dep in step.dependencies}

        # handle scan aliases: FROM foo AS bar
        source = getattr(step, "source", None)
        alias = getattr(source, "args", {}).get("alias")
        if alias is not None and (source_name := self.get(source.name)) is not None:
            self[alias.name] = source_name

        return Catalog({**self, **updates})


@singledispatch
def convert(step, catalog):
    raise TypeError(type(step))


@convert.register(sgp.Scan)
def convert_scan(scan, catalog):
    catalog = catalog.overlay(scan)

    table = catalog[scan.source.alias_or_name]

    if scan.condition:
        pred = convert(scan.condition, catalog=catalog)
        table = table.filter(pred)

    if scan.projections:
        projs = [convert(proj, catalog=catalog) for proj in scan.projections]
        table = table.select(projs)

    if isinstance(scan.limit, int):
        table = table.limit(scan.limit)

    return table


@convert.register(sgp.Sort)
def convert_sort(sort, catalog):
    catalog = catalog.overlay(sort)

    table = catalog[sort.name]

    if sort.key:
        keys = [convert(key, catalog=catalog) for key in sort.key]
        table = table.order_by(keys)

    if sort.projections:
        projs = [convert(proj, catalog=catalog) for proj in sort.projections]
        table = table.select(projs)

    return table


_join_types = {
    "": "inner",
    "LEFT": "left",
    "RIGHT": "right",
}


@convert.register(sgp.Join)
def convert_join(join, catalog):
    catalog = catalog.overlay(join)

    left_name = join.name
    left_table = catalog[left_name]

    for right_name, desc in join.joins.items():
        right_table = catalog[right_name]
        join_kind = _join_types[desc["side"]]

        if desc["join_key"]:
            predicate = None
            for left_key, right_key in zip(desc["source_key"], desc["join_key"]):
                left_key = convert(left_key, catalog=catalog)
                right_key = convert(right_key, catalog=catalog)
                if predicate is None:
                    predicate = left_key == right_key
                else:
                    predicate &= left_key == right_key
        else:
            condition = desc["condition"]
            predicate = convert(condition, catalog=catalog)

        left_table = left_table.join(right_table, predicates=predicate, how=join_kind)

    if join.condition:
        predicate = convert(join.condition, catalog=catalog)
        left_table = left_table.filter(predicate)

    catalog[left_name] = left_table

    return left_table


@convert.register(sgp.Aggregate)
def convert_aggregate(agg, catalog):
    catalog = catalog.overlay(agg)

    table = catalog[agg.source]
    if agg.aggregations:
        metrics = [convert(a, catalog=catalog) for a in agg.aggregations]
        groups = [convert(g, catalog=catalog) for k, g in agg.group.items()]
        table = table.aggregate(metrics, by=groups)

    return table


@convert.register(sge.Subquery)
def convert_subquery(subquery, catalog):
    tree = sgo.optimize(subquery.this, catalog.to_sqlglot(), rules=sgo.RULES)
    plan = sgp.Plan(tree)
    return convert(plan.root, catalog=catalog)


@convert.register(sge.Literal)
def convert_literal(literal, catalog):
    value = literal.this
    if literal.is_int:
        value = int(value)
    elif literal.is_number:
        value = float(value)
    return ibis.literal(value)


@convert.register(sge.Boolean)
def convert_boolean(boolean, catalog):
    return ibis.literal(boolean.this)


@convert.register(sge.Alias)
def convert_alias(alias, catalog):
    this = convert(alias.this, catalog=catalog)
    return this.name(alias.alias)


@convert.register(sge.Column)
def convert_column(column, catalog):
    table = catalog[column.table]
    return table[column.name]


@convert.register(sge.Ordered)
def convert_ordered(ordered, catalog):
    this = convert(ordered.this, catalog=catalog)
    desc = ordered.args["desc"]  # not exposed as an attribute
    return ibis.desc(this) if desc else ibis.asc(this)


_unary_operations = {
    sge.Paren: lambda x: x,
}


@convert.register(sge.Unary)
def convert_unary(unary, catalog):
    op = _unary_operations[type(unary)]
    this = convert(unary.this, catalog=catalog)
    return op(this)


_binary_operations = {
    sge.LT: operator.lt,
    sge.LTE: operator.le,
    sge.GT: operator.gt,
    sge.GTE: operator.ge,
    sge.EQ: operator.eq,
    sge.NEQ: operator.ne,
    sge.Add: operator.add,
    sge.Sub: operator.sub,
    sge.Mul: operator.mul,
    sge.Div: operator.truediv,
    sge.Pow: operator.pow,
    sge.And: operator.and_,
    sge.Or: operator.or_,
}


@convert.register(sge.Binary)
def convert_binary(binary, catalog):
    op = _binary_operations[type(binary)]
    this = convert(binary.this, catalog=catalog)
    expr = convert(binary.expression, catalog=catalog)

    if isinstance(binary.expression, sge.Subquery):
        # expr is a table expression
        assert len(expr.columns) == 1
        name = expr.columns[0]
        expr = expr[name]

    return op(this, expr)


_reduction_methods = {
    sge.Max: "max",
    sge.Min: "min",
    sge.Quantile: "quantile",
    sge.Sum: "sum",
    sge.Avg: "mean",
    sge.Count: "count",
}


@convert.register(sge.AggFunc)
def convert_sum(reduction, catalog):
    method = _reduction_methods[type(reduction)]
    this = convert(reduction.this, catalog=catalog)
    return getattr(this, method)()


@convert.register(sge.In)
def convert_in(in_, catalog):
    this = convert(in_.this, catalog=catalog)
    candidates = [convert(expression, catalog) for expression in in_.expressions]
    return this.isin(candidates)


@public
@experimental
def parse_sql(sqlstring, catalog, dialect=None):
    """Parse a SQL string into an Ibis expression.

    Parameters
    ----------
    sqlstring : str
        SQL string to parse
    catalog : dict
        A dictionary mapping table names to either schemas or ibis table expressions.
        If a schema is passed, a table expression will be created using the schema.
    dialect : str, optional
        The SQL dialect to use with sqlglot to parse the query string.

    Returns
    -------
    expr : ir.Expr

    """
    catalog = Catalog(
        {name: ibis.table(schema, name=name) for name, schema in catalog.items()}
    )

    expr = sg.parse_one(sqlstring, dialect)
    tree = sgo.optimize(expr, catalog.to_sqlglot(), rules=sgo.RULES)
    plan = sgp.Plan(tree)

    return convert(plan.root, catalog=catalog)


class SQLString(str):
    """Object to hold a formatted SQL string.

    Syntax highlights in Jupyter notebooks.
    """

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({str(self)!r})"

    def _repr_markdown_(self) -> str:
        return f"```sql\n{self!s}\n```"

    def _repr_pretty_(self, p, cycle) -> str:
        output = str(self)
        try:
            from pygments import highlight
            from pygments.formatters import TerminalFormatter
            from pygments.lexers import SqlLexer
        except ImportError:
            pass
        else:
            with contextlib.suppress(Exception):
                output = highlight(
                    code=output,
                    lexer=SqlLexer(),
                    formatter=TerminalFormatter(),
                )

        # strip trailing newline
        p.text(output.strip())


@public
def to_sql(expr: ir.Expr, dialect: str | None = None, **kwargs) -> SQLString:
    """Return the formatted SQL string for an expression.

    Parameters
    ----------
    expr
        Ibis expression.
    dialect
        SQL dialect to use for compilation.
    kwargs
        Scalar parameters

    Returns
    -------
    str
        Formatted SQL string

    """
    # try to infer from a non-str expression or if not possible fallback to
    # the default pretty dialect for expressions
    if dialect is None:
        try:
            backend = expr._find_backend()
        except com.IbisError:
            # default to duckdb for SQL compilation because it supports the
            # widest array of ibis features for SQL backends
            backend = ibis.duckdb
            read = "duckdb"
            write = ibis.options.sql.default_dialect
        else:
            read = write = backend.dialect
    else:
        try:
            backend = getattr(ibis, dialect)
        except AttributeError:
            raise ValueError(f"Unknown dialect {dialect}")
        else:
            read = write = getattr(backend, "dialect", dialect)

    sql = backend._to_sql(expr.unbind(), **kwargs)
    (pretty,) = sg.transpile(sql, read=read, write=write, pretty=True)
    return SQLString(pretty)



"""Tests for geo spatial data types."""
from __future__ import annotations

import numpy as np
import pytest
from numpy import testing
from pytest import param

import ibis
import ibis.expr.datatypes as dt

pytestmark = pytest.mark.geospatial


# TODO find a way to just run for the backends that support geo, without
# skipping if dependencies are missing
pytest.importorskip("geopandas")
shapely = pytest.importorskip("shapely")

# geo literals declaration
point_0 = ibis.literal((0, 0), type="point").name("p")
point_0_4326 = ibis.literal((0, 0), type="point;4326").name("p")

point_geom_0 = ibis.literal((0, 0), type="point;4326:geometry").name("p")
point_geom_1 = ibis.literal((1, 1), type="point;4326:geometry").name("p")
point_geom_0_srid0 = ibis.literal((0, 0), type="point;0:geometry").name("p")
point_geom_1_srid0 = ibis.literal((1, 1), type="point;0:geometry").name("p")
point_geom_2 = ibis.literal((2, 2), type="point;4326:geometry").name("p")
point_geog_0 = ibis.literal((0, 0), type="point;4326:geography").name("p")
point_geog_1 = ibis.literal((1, 1), type="point;4326:geography").name("p")
point_geog_2 = ibis.literal((2, 2), type="point;4326:geography").name("p")
polygon_0 = ibis.literal(
    (
        ((1, 0), (0, 1), (-1, 0), (0, -1), (1, 0)),
        ((0.1, 0), (0, 0.1), (-0.1, 0), (0, -0.1), (0.1, 0)),
    ),
    type="polygon;4326:geometry",
).name("p")

# test input data with shapely geometries
shp_point_0 = shapely.geometry.Point(0, 0)
shp_point_1 = shapely.geometry.Point(1, 1)
shp_point_2 = shapely.geometry.Point(2, 2)

shp_linestring_0 = shapely.geometry.LineString([shp_point_0, shp_point_1, shp_point_2])
shp_linestring_1 = shapely.geometry.LineString([shp_point_2, shp_point_1, shp_point_0])
shp_polygon_0 = shapely.geometry.Polygon(shp_linestring_0)
shp_multilinestring_0 = shapely.geometry.MultiLineString(
    [shp_linestring_0, shp_linestring_1]
)
shp_multipoint_0 = shapely.geometry.MultiPoint([shp_point_0, shp_point_1, shp_point_2])
shp_multipolygon_0 = shapely.geometry.MultiPolygon([shp_polygon_0])


@pytest.mark.parametrize(
    "expr",
    [
        point_0,
        point_0_4326,
        point_geom_0,
        point_geom_1,
        point_geom_2,
        point_geog_0,
        point_geog_1,
        point_geog_2,
    ],
)
def test_literal_geospatial_explicit(con, expr, snapshot):
    result = con.compile(expr)
    snapshot.assert_match(result, "out.sql")


@pytest.mark.parametrize(
    "shp",
    [
        shp_point_0,
        shp_point_1,
        shp_point_2,
        shp_linestring_0,
        shp_linestring_1,
        shp_polygon_0,
        shp_multipolygon_0,
        shp_multilinestring_0,
        shp_multipoint_0,
    ],
)
def test_literal_geospatial_inferred(con, shp, snapshot):
    snapshot.assert_match(con.compile(ibis.literal(shp)), "out.sql")


@pytest.mark.parametrize(
    ("expr_fn", "expected"),
    [
        param(lambda t: t["geo_linestring"].length(), [1.41] * 5, id="length"),
        param(lambda t: t["geo_point"].x(), [0, 1, 2, 3, 4], id="x"),
        param(lambda t: t["geo_point"].y(), [0, 1, 2, 3, 4], id="y"),
        param(
            lambda t: t["geo_linestring"].x_min(),
            [0, 1, 2, 3, 4],
            id="x_min",
            marks=pytest.mark.notimpl(["postgres"]),
        ),
        param(
            lambda t: t["geo_linestring"].x_max(),
            [1, 2, 3, 4, 5],
            id="x_max",
            marks=pytest.mark.notimpl(["postgres"]),
        ),
        param(
            lambda t: t["geo_linestring"].y_min(),
            [0, 1, 2, 3, 4],
            id="y_min",
            marks=pytest.mark.notimpl(["postgres"]),
        ),
        param(
            lambda t: t["geo_linestring"].y_max(),
            [1, 2, 3, 4, 5],
            id="y_max",
            marks=pytest.mark.notimpl(["postgres"]),
        ),
        param(
            lambda t: t["geo_multipolygon"].n_rings(),
            [2, 3, 1, 1, 1],
            id="n_rings",
            marks=pytest.mark.notimpl(["postgres"]),
        ),
        param(
            lambda t: t["geo_polygon"].set_srid(4326).perimeter(),
            [96.34, 114.36, 10.24, 10.24, 10.24],
            id="perimeter",
            marks=pytest.mark.notimpl(
                ["postgres"], reason="TODO: fix different results issue"
            ),
        ),
        param(
            lambda t: t["geo_multipolygon"].n_points(),
            [7, 11, 5, 5, 5],
            id="n_points",
            marks=pytest.mark.notimpl(
                ["postgres"], reason="TODO: fix different results issue"
            ),
        ),
    ],
)
def test_geo_spatial_unops(geotable, expr_fn, expected):
    """Testing for geo spatial unary operations."""
    expr = expr_fn(geotable)
    result = expr.execute()
    testing.assert_almost_equal(result, expected, decimal=2)


@pytest.mark.parametrize(
    ("expr_fn", "expected"),
    [
        param(
            lambda t: t["geo_linestring"].contains(point_geom_1_srid0),
            [False] * 5,  # does not contain the border
            id="contains",
        ),
        param(
            lambda t: t["geo_linestring"].disjoint(point_geom_0_srid0),
            [False, True, True, True, True],
            id="disjoint",
        ),
        param(
            lambda t: t["geo_point"].d_within(point_geom_1_srid0, 2.0),
            [True, True, True, False, False],
            id="d_within",
        ),
        param(
            lambda t: t["geo_point"].d_fully_within(t["geo_linestring"], 2.0),
            [True, True, True, True, True],
            id="d_fully_within",
        ),
        param(
            lambda t: t["geo_linestring"].intersects(point_geom_0_srid0),
            [True, False, False, False, False],
            id="intersects",
        ),
        param(
            lambda t: t["geo_linestring"].distance(point_geom_0_srid0),
            [0.0, 1.41, 2.82, 4.24, 5.66],
            id="distance",
        ),
        param(
            lambda t: t["geo_linestring"].max_distance(point_geom_0_srid0),
            [1.41, 2.82, 4.24, 5.66, 7.08],
            id="max_distance",
            marks=pytest.mark.notimpl(["postgres"]),
        ),
        param(
            lambda t: t.geo_polygon.contains(ibis.literal(30).point(10)),
            [True, False, False, False, False],
            id="point",
            marks=pytest.mark.notimpl(["postgres"]),
        ),
    ],
)
def test_geo_spatial_binops(geotable, expr_fn, expected):
    """Testing for geo spatial binary operations."""
    expr = expr_fn(geotable)
    result = expr.execute()
    testing.assert_almost_equal(result, expected, decimal=2)


@pytest.mark.parametrize(
    ("expr_fn", "expected"),
    [
        param(
            lambda t: t["geo_linestring"].end_point(),
            [True, True, True, True, True],
            id="end_point",
        ),
        param(
            lambda t: t["geo_linestring"].point_n(1),
            [True, True, True, True, True],
            id="point_n",
            marks=pytest.mark.notimpl(["postgres"]),
        ),
        param(
            lambda t: t["geo_linestring"].start_point(),
            [True, True, True, True, True],
            id="start_point",
        ),
    ],
)
def test_get_point(geotable, expr_fn, expected):
    """Testing for geo spatial get point operations."""
    arg = expr_fn(geotable)
    # NB: there is a difference in how OmnisciDB and PostGIS consider
    # boundaries with the contains predicate. Work around this by adding a
    # small buffer.
    expr = geotable["geo_linestring"].buffer(0.01).contains(arg)
    result = geotable[geotable, expr.name("tmp")].execute()["tmp"]
    testing.assert_almost_equal(result, expected, decimal=2)


def test_area(con, geotable):
    """Testing for geo spatial area operation."""
    expr = geotable.select("id", tmp=polygon_0.area())
    result = expr.execute()["tmp"].values
    expected = np.array([con.execute(polygon_0).area] * len(result))
    assert pytest.approx(result) == expected


@pytest.mark.parametrize(
    ("condition", "expected"),
    [
        (lambda _: point_geom_2.srid(), 4326),
        (lambda _: point_geom_0.srid(), 4326),
        (lambda t: t.geo_point.srid(), 0),
        (lambda t: t.geo_linestring.srid(), 0),
        (lambda t: t.geo_polygon.srid(), 0),
        (lambda t: t.geo_multipolygon.srid(), 0),
    ],
)
def test_srid(geotable, condition, expected):
    """Testing for geo spatial srid operation."""
    expr = geotable[geotable.id, condition(geotable).name("tmp")]
    result = expr.execute()["tmp"][[0]]
    assert np.all(result == expected)


@pytest.mark.parametrize(
    ("condition", "expected"),
    [
        (lambda _: point_geom_0.set_srid(4326).srid(), 4326),
        (lambda _: point_geom_0.set_srid(4326).set_srid(0).srid(), 0),
        (lambda t: t.geo_point.set_srid(4326).srid(), 4326),
        (lambda t: t.geo_linestring.set_srid(4326).srid(), 4326),
        (lambda t: t.geo_polygon.set_srid(4326).srid(), 4326),
        (lambda t: t.geo_multipolygon.set_srid(4326).srid(), 4326),
    ],
)
def test_set_srid(geotable, condition, expected):
    """Testing for geo spatial set_srid operation."""
    expr = geotable[geotable.id, condition(geotable).name("tmp")]
    result = expr.execute()["tmp"][[0]]
    assert np.all(result == expected)


@pytest.mark.parametrize(
    ("condition", "expected"),
    [
        (lambda t: point_geom_0.transform(900913).srid(), 900913),
        (lambda t: point_geom_2.transform(900913).srid(), 900913),
        (
            lambda t: t.geo_point.set_srid(4326).transform(900913).srid(),
            900913,
        ),
        (
            lambda t: t.geo_linestring.set_srid(4326).transform(900913).srid(),
            900913,
        ),
        (
            lambda t: t.geo_polygon.set_srid(4326).transform(900913).srid(),
            900913,
        ),
        (
            lambda t: t.geo_multipolygon.set_srid(4326).transform(900913).srid(),
            900913,
        ),
    ],
)
def test_transform(geotable, condition, expected):
    """Testing for geo spatial transform operation."""
    expr = geotable[geotable.id, condition(geotable).name("tmp")]
    result = expr.execute()["tmp"][[0]]
    assert np.all(result == expected)


@pytest.mark.parametrize(
    "expr_fn",
    [
        param(lambda t: t.geo_point.set_srid(4326), id="geom_geo_point"),
        param(lambda t: point_geom_0, id="point_geom_0"),
        param(lambda t: point_geom_1, id="point_geom_1"),
        param(lambda t: point_geom_2, id="point_geom_2"),
        param(lambda t: point_geog_0, id="point_geog_0"),
        param(lambda t: point_geog_1, id="point_geog_1"),
        param(lambda t: point_geog_2, id="point_geog_2"),
    ],
)
def test_cast_geography(geotable, expr_fn):
    """Testing for geo spatial transform operation."""
    p = expr_fn(geotable).cast("geography")
    expr = geotable[geotable.id, p.distance(p).name("tmp")]
    result = expr.execute()["tmp"][[0]]
    # distance from a point to a same point should be 0
    assert np.all(result == 0)


@pytest.mark.parametrize(
    "expr_fn",
    [
        param(lambda t: t.geo_point.set_srid(4326), id="t_geo_point"),
        param(lambda _: point_geom_0, id="point_geom_0"),
        param(lambda _: point_geom_1, id="point_geom_1"),
        param(lambda _: point_geom_2, id="point_geom_2"),
        param(lambda _: point_geog_0, id="point_geog_0"),
        param(lambda _: point_geog_1, id="point_geog_1"),
        param(lambda _: point_geog_2, id="point_geog_2"),
    ],
)
def test_cast_geometry(geotable, expr_fn):
    """Testing for geo spatial transform operation."""
    p = expr_fn(geotable).cast("geometry")
    expr = geotable[geotable.id, p.distance(p).name("tmp")]
    result = expr.execute()["tmp"][[0]]
    # distance from a point to a same point should be 0
    assert np.all(result == 0)


def test_geo_dataframe(geotable):
    """Testing for geo dataframe output."""
    import geopandas

    assert isinstance(geotable.execute(), geopandas.geodataframe.GeoDataFrame)


@pytest.mark.parametrize(
    "modifier",
    [
        param({}, id="none"),
        param({"srid": 4326}, id="srid"),
        param({"srid": 4326, "geotype": "geometry"}, id="geometry"),
        param({"srid": 4326, "geotype": "geography"}, id="geography"),
    ],
)
@pytest.mark.parametrize(
    ("shape", "value"),
    [
        # Geometry primitives (2D)
        param("point", (30, 10), id="point"),
        param("linestring", ((30, 10), (10, 30), (40, 40)), id="linestring"),
        param(
            "polygon",
            (
                ((35, 10), (45, 45), (15, 40), (10, 20), (35, 10)),
                ((20, 30), (35, 35), (30, 20), (20, 30)),
            ),
            id="polygon",
        ),
        param(
            "polygon",
            (((30, 10), (40, 40), (20, 40), (10, 20), (30, 10)),),
            id="polygon_single",
        ),
        # Multipart geometries (2D)
        param("multipoint", ((10, 40), (40, 30), (20, 20), (30, 10)), id="multipoint"),
        param(
            "multilinestring",
            (((10, 10), (20, 20), (10, 40)), ((40, 40), (30, 30), (40, 20), (30, 10))),
            id="multilinestring",
        ),
        param(
            "multipolygon",
            (
                (((40, 40), (20, 45), (45, 30), (40, 40)),),
                (
                    ((20, 35), (10, 30), (10, 10), (30, 5), (45, 20), (20, 35)),
                    ((30, 20), (20, 15), (20, 25), (30, 20)),
                ),
            ),
            id="multipolygon",
        ),
    ],
)
def test_geo_literals_smoke(con, shape, value, modifier, snapshot):
    """Smoke tests for geo spatial literals."""
    expr = ibis.literal(value, type=getattr(dt, shape).copy(**modifier))
    snapshot.assert_match(con.compile(expr), "out.sql")


@pytest.mark.parametrize(
    "fn_expr",
    [
        pytest.param(lambda t: t.geo_point.srid(), id="point_srid"),
        pytest.param(lambda t: t.geo_point.set_srid(4326), id="point_set_srid"),
        pytest.param(lambda t: t.geo_point.x(), id="point_x"),
        pytest.param(lambda t: t.geo_point.y(), id="point_y"),
        pytest.param(
            lambda t: t.geo_linestring.contains(t.geo_point),
            id="linestring_contains",
        ),
        pytest.param(lambda t: t.geo_linestring.end_point(), id="linestring_end_point"),
        pytest.param(lambda t: t.geo_linestring.length(), id="linestring_length"),
        pytest.param(
            lambda t: t.geo_linestring.max_distance(t.geo_point),
            id="linestring_max_distance",
            marks=pytest.mark.notimpl(["postgres"]),
        ),
        pytest.param(
            lambda t: t.geo_linestring.point_n(1),
            id="linestring_point_n",
            marks=pytest.mark.notimpl(["postgres"]),
        ),
        pytest.param(
            lambda t: t.geo_linestring.start_point(),
            id="linestring_start_point",
        ),
        pytest.param(
            lambda t: t.geo_linestring.x_max(),
            id="linestring_x_max",
            marks=pytest.mark.notimpl(["postgres"]),
        ),
        pytest.param(
            lambda t: t.geo_linestring.x_min(),
            id="linestring_x_min",
            marks=pytest.mark.notimpl(["postgres"]),
        ),
        pytest.param(
            lambda t: t.geo_linestring.y_max(),
            id="linestring_y_max",
            marks=pytest.mark.notimpl(["postgres"]),
        ),
        pytest.param(
            lambda t: t.geo_linestring.y_min(),
            id="linestring_y_min",
            marks=pytest.mark.notimpl(["postgres"]),
        ),
        pytest.param(lambda t: t.geo_polygon.area(), id="polygon_area"),
        pytest.param(lambda t: t.geo_polygon.perimeter(), id="polygon_perimeter"),
        pytest.param(
            lambda t: t.geo_multipolygon.n_points(), id="multipolygon_n_points"
        ),
        pytest.param(
            lambda t: t.geo_multipolygon.n_rings(),
            id="multipolygon_n_rings",
            marks=pytest.mark.notimpl(["postgres"]),
        ),
    ],
)
def test_geo_ops_smoke(geotable, fn_expr, snapshot):
    """Smoke tests for geo spatial operations."""
    snapshot.assert_match(fn_expr(geotable).compile(), "out.sql")


def test_geo_equals(geotable, snapshot):
    # Fix https://github.com/ibis-project/ibis/pull/2956
    expr = geotable.mutate(
        Location_Latitude=geotable.geo_point.y(), Latitude=geotable.geo_point.y()
    )

    snapshot.assert_match(expr.compile(), "out1.sql")

    # simple test using ==
    expr = geotable.geo_point == geotable.geo_point
    snapshot.assert_match(expr.compile(), "out2.sql")
    result = expr.execute()
    assert not result.empty
    assert result.all()

    # using geo_equals
    expr = geotable.geo_point.geo_equals(geotable.geo_point).name("tmp")
    snapshot.assert_match(expr.compile(), "out3.sql")
    result = expr.execute()
    assert not result.empty
    assert result.all()

    # equals returns a boolean object
    assert geotable.geo_point.equals(geotable.geo_point)



from __future__ import annotations

import ibis

from .conftest import add_date, tpch_test


@tpch_test
def test_tpc_h04(orders, lineitem):
    """Order Priority Checking Query (Q4)"""
    DATE = "1993-07-01"
    cond = (lineitem.l_orderkey == orders.o_orderkey) & (
        lineitem.l_commitdate < lineitem.l_receiptdate
    )
    q = orders.filter(
        [
            cond.any(),
            orders.o_orderdate >= ibis.date(DATE),
            orders.o_orderdate < add_date(DATE, dm=3),
        ]
    )
    q = q.group_by([orders.o_orderpriority])
    q = q.aggregate(order_count=lambda t: t.count())
    q = q.order_by([orders.o_orderpriority])
    return q



import ibis


star1 = ibis.table(
    name="star1",
    schema={"c": "int32", "f": "float64", "foo_id": "string", "bar_id": "string"},
)
star2 = ibis.table(
    name="star2", schema={"foo_id": "string", "value1": "float64", "value3": "float64"}
)

result = star1.semi_join(star2, star1.foo_id == star2.foo_id).select(
    star1.c, star1.f, star1.foo_id, star1.bar_id
)



import ibis


star2 = ibis.table(
    name="star2", schema={"foo_id": "string", "value1": "float64", "value3": "float64"}
)
star1 = ibis.table(
    name="star1",
    schema={"c": "int32", "f": "float64", "foo_id": "string", "bar_id": "string"},
)
agg = star1.group_by(star1.foo_id).aggregate(star1.f.sum().name("total"))
proj = agg.inner_join(star2, agg.foo_id == star2.foo_id).select([agg, star2.value1])
proj1 = proj.filter(proj.total > 100)

result = proj1.order_by(proj1.total.desc())



from __future__ import annotations

import datetime

import pytest

import ibis
from ibis.backends.tests.sql.conftest import to_sql
from ibis.tests.util import assert_decompile_roundtrip, schemas_eq

pytestmark = pytest.mark.duckdb


def test_union(union, snapshot):
    snapshot.assert_match(to_sql(union), "out.sql")
    assert_decompile_roundtrip(union, snapshot, eq=schemas_eq)


def test_union_project_column(union_all, snapshot):
    # select a column, get a subquery
    expr = union_all[[union_all.key]]
    snapshot.assert_match(to_sql(expr), "out.sql")
    assert_decompile_roundtrip(expr, snapshot, eq=schemas_eq)


def test_table_intersect(intersect, snapshot):
    snapshot.assert_match(to_sql(intersect), "out.sql")
    assert_decompile_roundtrip(intersect, snapshot, eq=schemas_eq)


def test_table_difference(difference, snapshot):
    snapshot.assert_match(to_sql(difference), "out.sql")
    assert_decompile_roundtrip(difference, snapshot, eq=schemas_eq)


def test_intersect_project_column(intersect, snapshot):
    # select a column, get a subquery
    expr = intersect[[intersect.key]]
    snapshot.assert_match(to_sql(expr), "out.sql")
    assert_decompile_roundtrip(expr, snapshot, eq=schemas_eq)


def test_difference_project_column(difference, snapshot):
    # select a column, get a subquery
    expr = difference[[difference.key]]
    snapshot.assert_match(to_sql(expr), "out.sql")
    assert_decompile_roundtrip(expr, snapshot, eq=schemas_eq)


def test_table_distinct(con, snapshot):
    t = con.table("functional_alltypes")

    expr = t[t.string_col, t.int_col].distinct()
    snapshot.assert_match(to_sql(expr), "out.sql")
    assert_decompile_roundtrip(expr, snapshot)


def test_column_distinct(con, snapshot):
    t = con.table("functional_alltypes")
    expr = t[t.string_col].distinct()
    snapshot.assert_match(to_sql(expr), "out.sql")
    assert_decompile_roundtrip(expr, snapshot)


def test_count_distinct(con, snapshot):
    t = con.table("functional_alltypes")

    metric = t.int_col.nunique().name("nunique")
    expr = t[t.bigint_col > 0].group_by("string_col").aggregate([metric])
    snapshot.assert_match(to_sql(expr), "out.sql")
    assert_decompile_roundtrip(expr, snapshot)


def test_multiple_count_distinct(con, snapshot):
    # Impala and some other databases will not execute multiple
    # count-distincts in a single aggregation query. This error reporting
    # will be left to the database itself, for now.
    t = con.table("functional_alltypes")
    metrics = [
        t.int_col.nunique().name("int_card"),
        t.smallint_col.nunique().name("smallint_card"),
    ]

    expr = t.group_by("string_col").aggregate(metrics)
    snapshot.assert_match(to_sql(expr), "out.sql")
    assert_decompile_roundtrip(expr, snapshot)


def test_pushdown_with_or(snapshot):
    t = ibis.table(
        [
            ("double_col", "float64"),
            ("string_col", "string"),
            ("int_col", "int32"),
            ("float_col", "float32"),
        ],
        "functional_alltypes",
    )
    subset = t[(t.double_col > 3.14) & t.string_col.contains("foo")]
    expr = subset[(subset.int_col - 1 == 0) | (subset.float_col <= 1.34)]
    snapshot.assert_match(to_sql(expr), "out.sql")


def test_having_size(snapshot):
    t = ibis.table(
        [
            ("double_col", "double"),
            ("string_col", "string"),
            ("int_col", "int32"),
            ("float_col", "float"),
        ],
        "functional_alltypes",
    )
    expr = t.group_by(t.string_col).having(t.double_col.max() == 1).size()
    snapshot.assert_match(to_sql(expr), "out.sql")


def test_having_from_filter(snapshot):
    t = ibis.table([("a", "int64"), ("b", "string")], "t")
    filt = t[t.b == "m"]
    gb = filt.group_by(filt.b)
    having = gb.having(filt.a.max() == 2)
    expr = having.aggregate(filt.a.sum().name("sum"))
    snapshot.assert_match(to_sql(expr), "out.sql")
    # params get different auto incremented counter identifiers
    assert_decompile_roundtrip(expr, snapshot, eq=schemas_eq)


def test_simple_agg_filter(snapshot):
    t = ibis.table([("a", "int64"), ("b", "string")], name="my_table")
    filt = t[t.a < 100]
    expr = filt[filt.a == filt.a.max()]
    snapshot.assert_match(to_sql(expr), "out.sql")


def test_agg_and_non_agg_filter(snapshot):
    t = ibis.table([("a", "int64"), ("b", "string")], name="my_table")
    filt = t[t.a < 100]
    expr = filt[filt.a == filt.a.max()]
    expr = expr[expr.b == "a"]
    snapshot.assert_match(to_sql(expr), "out.sql")


def test_agg_filter(snapshot):
    t = ibis.table([("a", "int64"), ("b", "int64")], name="my_table")
    t = t.mutate(b2=t.b * 2)
    t = t[["a", "b2"]]
    filt = t[t.a < 100]
    expr = filt[filt.a == filt.a.max().name("blah")]
    snapshot.assert_match(to_sql(expr), "out.sql")


def test_agg_filter_with_alias(snapshot):
    t = ibis.table([("a", "int64"), ("b", "int64")], name="my_table")
    t = t.mutate(b2=t.b * 2)
    t = t[["a", "b2"]]
    filt = t[t.a < 100]
    expr = filt[filt.a.name("A") == filt.a.max().name("blah")]
    snapshot.assert_match(to_sql(expr), "out.sql")


def test_table_drop_with_filter(snapshot):
    left = ibis.table(
        [("a", "int64"), ("b", "string"), ("c", "timestamp")], name="t"
    ).rename(C="c")
    left = left.filter(left.C == datetime.datetime(2018, 1, 1))
    left = left.drop("C")
    left = left.mutate(the_date=datetime.datetime(2018, 1, 1))

    right = ibis.table([("b", "string")], name="s")
    joined = left.join(right, left.b == right.b)
    joined = joined[left.a]
    expr = joined.filter(joined.a < 1.0)
    snapshot.assert_match(to_sql(expr), "out.sql")
    assert_decompile_roundtrip(expr, snapshot, eq=schemas_eq)


def test_table_drop_consistency():
    # GH2829
    t = ibis.table([("a", "int64"), ("b", "string"), ("c", "timestamp")], name="t")

    expected = t.select(["a", "c"])
    result = t.drop("b")

    assert expected.schema() == result.schema()
    assert set(result.columns) == {"a", "c"}


def test_subquery_where_location(snapshot):
    t = ibis.table(
        [
            ("float_col", "float32"),
            ("timestamp_col", "timestamp"),
            ("int_col", "int32"),
            ("string_col", "string"),
        ],
        name="alltypes",
    )
    param = ibis.param("timestamp").name("my_param")
    expr = (
        t[["float_col", "timestamp_col", "int_col", "string_col"]][
            lambda t: t.timestamp_col < param
        ]
        .group_by("string_col")
        .aggregate(foo=lambda t: t.float_col.sum())
        .foo.count()
    )
    out = to_sql(expr, params={param: "20140101"})
    snapshot.assert_match(out, "out.sql")
    # params get different auto incremented counter identifiers
    assert_decompile_roundtrip(expr, snapshot, eq=schemas_eq)


def test_column_expr_retains_name(snapshot):
    t = ibis.table([("int_col", "int32")], name="int_col_table")
    expr = (t.int_col + 4).name("foo")
    snapshot.assert_match(to_sql(expr), "out.sql")
    assert_decompile_roundtrip(expr, snapshot)


def test_column_expr_default_name(snapshot):
    t = ibis.table([("int_col", "int32")], name="int_col_table")
    expr = t.int_col + 4
    snapshot.assert_match(to_sql(expr), "out.sql")
    assert_decompile_roundtrip(expr, snapshot)


def test_union_order_by(snapshot):
    t = ibis.table(dict(a="int", b="string"), name="t")
    expr = t.order_by("b").union(t.order_by("b"))
    snapshot.assert_match(to_sql(expr), "out.sql")
    assert_decompile_roundtrip(expr, snapshot, eq=schemas_eq)



import ibis


employee = ibis.table(
    name="employee",
    schema={"first_name": "string", "last_name": "string", "id": "int64"},
)
f = employee.filter(
    employee.first_name.isin(
        (
            ibis.literal("Graham"),
            ibis.literal("John"),
            ibis.literal("Terry"),
            ibis.literal("Eric"),
            ibis.literal("Michael"),
        )
    )
)

result = f.select(f.first_name)



from __future__ import annotations

import ibis

from .conftest import add_date, tpch_test


@tpch_test
def test_tpc_h20(supplier, nation, partsupp, part, lineitem):
    """Potential Part Promotion Query (Q20)

    The Potential Part Promotion Query identifies suppliers in a particular
    nation having selected parts that may be candidates for a promotional
    offer."""
    COLOR = "forest"
    DATE = "1994-01-01"
    NATION = "CANADA"

    q1 = supplier.join(nation, supplier.s_nationkey == nation.n_nationkey)

    q3 = part.filter([part.p_name.like(f"{COLOR}%")])
    q2 = partsupp

    q4 = lineitem.filter(
        [
            lineitem.l_partkey == q2.ps_partkey,
            lineitem.l_suppkey == q2.ps_suppkey,
            lineitem.l_shipdate >= ibis.date(DATE),
            lineitem.l_shipdate < add_date(DATE, dy=1),
        ]
    )

    q2 = q2.filter(
        [
            partsupp.ps_partkey.isin(q3.p_partkey),
            partsupp.ps_availqty > 0.5 * q4.l_quantity.sum(),
        ]
    )

    q1 = q1.filter([q1.n_name == NATION, q1.s_suppkey.isin(q2.ps_suppkey)])

    q1 = q1[q1.s_name, q1.s_address]

    return q1.order_by(q1.s_name)



from __future__ import annotations

import re

import dask.dataframe as dd
import numpy as np
import pandas as pd
import pytest
from dask.dataframe.utils import tm
from pytest import param

import ibis
import ibis.expr.operations as ops


def make_dask_data_frame(npartitions):
    df = pd.DataFrame(np.random.randn(30, 4), columns=list("ABCD"))
    return dd.from_pandas(df, npartitions=npartitions)


@pytest.fixture
def client(npartitions):
    return ibis.dask.connect(
        {
            "df": dd.from_pandas(
                pd.DataFrame({"a": [1, 2, 3], "b": list("abc")}),
                npartitions=npartitions,
            ),
            "df_unknown": dd.from_pandas(
                pd.DataFrame({"array_of_strings": [["a", "b"], [], ["c"]]}),
                npartitions=npartitions,
            ),
        }
    )


@pytest.fixture
def table(client):
    return client.table("df")


def test_connect_no_args():
    con = ibis.dask.connect()
    assert dict(con.tables) == {}


def test_client_table(table):
    assert isinstance(table.op(), ops.DatabaseTable)


def test_create_table(client, npartitions):
    ddf = make_dask_data_frame(npartitions)
    client.create_table("testing", obj=ddf)
    assert "testing" in client.list_tables()
    client.create_table("testingschema", schema=client.get_schema("testing"))
    assert "testingschema" in client.list_tables()


def test_literal(client):
    lit = ibis.literal(1)
    result = client.execute(lit)
    assert result == 1


def test_list_tables(client):
    assert client.list_tables(like="df_unknown")
    assert not client.list_tables(like="not_in_the_database")
    assert client.list_tables()


def test_drop(table):
    table = table.mutate(c=table.a)
    expr = table.drop("a")
    result = expr.execute()
    expected = table[["b", "c"]].execute()
    tm.assert_frame_equal(result, expected)


@pytest.mark.parametrize(
    "unit",
    [
        "Y",
        "M",
        "D",
        "h",
        "m",
        "s",
        "ms",
        "us",
        "ns",
        param("ps", marks=pytest.mark.xfail),
        param("fs", marks=pytest.mark.xfail),
        param("as", marks=pytest.mark.xfail),
    ],
)
def test_datetime64_infer(client, unit):
    value = np.datetime64("2018-01-02", unit)
    expr = ibis.literal(value, type="timestamp")
    result = client.execute(expr)
    assert result == pd.Timestamp(value).to_pydatetime()


def test_invalid_connection_parameter_types(npartitions):
    # Check that the user receives a TypeError with an informative message when
    # passing invalid an connection parameter to the backend.
    expected_msg = re.escape(
        "Expected an instance of 'dask.dataframe.DataFrame' for 'invalid_str',"
        " got an instance of 'str' instead."
    )
    with pytest.raises(TypeError, match=expected_msg):
        ibis.dask.connect(
            {
                "valid_dask_df": dd.from_pandas(
                    pd.DataFrame({"a": [1, 2, 3], "b": list("abc")}),
                    npartitions=npartitions,
                ),
                "valid_pandas_df": pd.DataFrame({"a": [1, 2, 3], "b": list("abc")}),
                "invalid_str": "file.csv",
            }
        )

    expected_msg = re.escape(
        "Expected an instance of 'dask.dataframe.DataFrame' for 'df', "
        "got an instance of 'str' instead."
    )
    con = ibis.dask.connect()
    with pytest.raises(TypeError, match=expected_msg):
        con.from_dataframe("file.csv")



from __future__ import annotations

import pytest

from ibis import literal as L
from ibis.backends.impala.tests.conftest import translate


@pytest.fixture(scope="module")
def table(mockcon):
    return mockcon.table("alltypes")


@pytest.mark.parametrize("method_name", ["isin", "notin"])
def test_field_in_literals(table, method_name, snapshot):
    values = ["foo", "bar", "baz"]
    method = getattr(table.g, method_name)
    expr = method(values)
    result = translate(expr)
    snapshot.assert_match(result, "out.sql")


@pytest.mark.parametrize("method_name", ["isin", "notin"])
def test_literal_in_fields(table, method_name, snapshot):
    values = [table.a, table.b, table.c]
    method = getattr(L(2), method_name)
    expr = method(values)
    result = translate(expr)
    snapshot.assert_match(result, "out.sql")


@pytest.mark.parametrize("method_name", ["isin", "notin"])
def test_isin_notin_in_select(table, method_name, snapshot):
    values = ["foo", "bar"]
    method = getattr(table.g, method_name)
    filtered = table[method(values)]
    result = translate(filtered)
    snapshot.assert_match(result, "out.sql")



from __future__ import annotations

from packaging.version import Version
from packaging.version import parse as parse_version

import ibis


def test_version():
    assert isinstance(parse_version(ibis.__version__), Version)



from __future__ import annotations

import hypothesis as h
import hypothesis.strategies as st
import pytest
from pytest import param

import ibis
import ibis.expr.datatypes as dt
import ibis.tests.strategies as its
from ibis.backends.base.sqlglot.datatypes import ClickHouseType

pytest.importorskip("clickhouse_connect")


def test_column_types(alltypes):
    df = alltypes.limit(1).execute()
    assert df.tinyint_col.dtype.name == "int8"
    assert df.smallint_col.dtype.name == "int16"
    assert df.int_col.dtype.name == "int32"
    assert df.bigint_col.dtype.name == "int64"
    assert df.float_col.dtype.name == "float32"
    assert df.double_col.dtype.name == "float64"
    assert df.timestamp_col.dtype.name == "datetime64[ns]"


def test_columns_types_with_additional_argument(con):
    sql_types = [
        "toFixedString('foo', 8) AS fixedstring_col",
        "toDateTime('2018-07-02 00:00:00', 'UTC') AS datetime_col",
        "toDateTime64('2018-07-02 00:00:00', 9, 'UTC') AS datetime_ns_col",
    ]
    df = con.sql(f"SELECT {', '.join(sql_types)}").execute()
    assert df.fixedstring_col.dtype.name == "object"
    assert df.datetime_col.dtype.name in ("datetime64[ns, UTC]", "datetime64[s, UTC]")
    assert df.datetime_ns_col.dtype.name == "datetime64[ns, UTC]"


def test_array_discovery_clickhouse(con):
    t = con.tables.array_types
    expected = ibis.schema(
        dict(
            x=dt.Array(dt.int64, nullable=False),
            y=dt.Array(dt.string, nullable=False),
            z=dt.Array(dt.float64, nullable=False),
            grouper=dt.string,
            scalar_column=dt.float64,
            multi_dim=dt.Array(dt.Array(dt.int64, nullable=False), nullable=False),
        )
    )
    assert t.schema() == expected


@pytest.mark.parametrize(
    ("ch_type", "ibis_type"),
    [
        param(
            "Enum8('' = 0, 'CDMA' = 1, 'GSM' = 2, 'LTE' = 3, 'NR' = 4)",
            dt.String(nullable=False),
            id="enum",
        ),
        param("IPv4", dt.inet(nullable=False), id="ipv4"),
        param("IPv6", dt.inet(nullable=False), id="ipv6"),
        param("JSON", dt.json(nullable=False), id="json"),
        param("Object('json')", dt.json(nullable=False), id="object_json"),
        param(
            "LowCardinality(String)", dt.String(nullable=False), id="low_card_string"
        ),
        param(
            "Array(Int8)",
            dt.Array(dt.Int8(nullable=False), nullable=False),
            id="array_int8",
        ),
        param(
            "Array(Int16)",
            dt.Array(dt.Int16(nullable=False), nullable=False),
            id="array_int16",
        ),
        param(
            "Array(Int32)",
            dt.Array(dt.Int32(nullable=False), nullable=False),
            id="array_int32",
        ),
        param(
            "Array(Int64)",
            dt.Array(dt.Int64(nullable=False), nullable=False),
            id="array_int64",
        ),
        param(
            "Array(UInt8)",
            dt.Array(dt.UInt8(nullable=False), nullable=False),
            id="array_uint8",
        ),
        param(
            "Array(UInt16)",
            dt.Array(dt.UInt16(nullable=False), nullable=False),
            id="array_uint16",
        ),
        param(
            "Array(UInt32)",
            dt.Array(dt.UInt32(nullable=False), nullable=False),
            id="array_uint32",
        ),
        param(
            "Array(UInt64)",
            dt.Array(dt.UInt64(nullable=False), nullable=False),
            id="array_uint64",
        ),
        param(
            "Array(Float32)",
            dt.Array(dt.Float32(nullable=False), nullable=False),
            id="array_float32",
        ),
        param(
            "Array(Float64)",
            dt.Array(dt.Float64(nullable=False), nullable=False),
            id="array_float64",
        ),
        param(
            "Array(String)",
            dt.Array(dt.String(nullable=False), nullable=False),
            id="array_string",
        ),
        param(
            "Array(FixedString(32))",
            dt.Array(dt.String(nullable=False), nullable=False),
            id="array_fixed_string",
        ),
        param(
            "Array(Date)",
            dt.Array(dt.Date(nullable=False), nullable=False),
            id="array_date",
        ),
        param(
            "Array(DateTime)",
            dt.Array(dt.Timestamp(scale=0, nullable=False), nullable=False),
            id="array_datetime",
        ),
        param(
            "Array(DateTime64(9))",
            dt.Array(dt.Timestamp(scale=9, nullable=False), nullable=False),
            id="array_datetime64",
        ),
        param("Array(Nothing)", dt.Array(dt.null, nullable=False), id="array_nothing"),
        param("Array(Null)", dt.Array(dt.null, nullable=False), id="array_null"),
        param(
            "Array(Array(Int8))",
            dt.Array(
                dt.Array(dt.Int8(nullable=False), nullable=False),
                nullable=False,
            ),
            id="double_array",
        ),
        param(
            "Array(Array(Array(Int8)))",
            dt.Array(
                dt.Array(
                    dt.Array(dt.Int8(nullable=False), nullable=False),
                    nullable=False,
                ),
                nullable=False,
            ),
            id="triple_array",
        ),
        param(
            "Array(Array(Array(Array(Int8))))",
            dt.Array(
                dt.Array(
                    dt.Array(
                        dt.Array(dt.Int8(nullable=False), nullable=False),
                        nullable=False,
                    ),
                    nullable=False,
                ),
                nullable=False,
            ),
            id="quad_array",
        ),
        param(
            "Map(Nullable(String), Nullable(UInt64))",
            dt.Map(dt.string, dt.uint64, nullable=False),
            id="map",
        ),
        param("Decimal(10, 3)", dt.Decimal(10, 3, nullable=False), id="decimal"),
        param(
            "Tuple(a String, b Array(Nullable(Float64)))",
            dt.Struct(
                dict(
                    a=dt.String(nullable=False),
                    b=dt.Array(dt.float64, nullable=False),
                ),
                nullable=False,
            ),
            id="named_tuple",
        ),
        param(
            "Tuple(String, Array(Nullable(Float64)))",
            dt.Struct(
                dict(
                    f0=dt.String(nullable=False),
                    f1=dt.Array(dt.float64, nullable=False),
                ),
                nullable=False,
            ),
            id="unnamed_tuple",
        ),
        param(
            "Tuple(a String, Array(Nullable(Float64)))",
            dt.Struct(
                dict(
                    a=dt.String(nullable=False),
                    f1=dt.Array(dt.float64, nullable=False),
                ),
                nullable=False,
            ),
            id="partially_named",
        ),
        param(
            "Nested(a String, b Array(Nullable(Float64)))",
            dt.Struct(
                dict(
                    a=dt.Array(dt.String(nullable=False), nullable=False),
                    b=dt.Array(dt.Array(dt.float64, nullable=False), nullable=False),
                ),
                nullable=False,
            ),
            id="nested",
        ),
        param("DateTime", dt.Timestamp(scale=0, nullable=False), id="datetime"),
        param(
            "DateTime('Europe/Budapest')",
            dt.Timestamp(scale=0, timezone="Europe/Budapest", nullable=False),
            id="datetime_timezone",
        ),
        param(
            "DateTime64(0)", dt.Timestamp(scale=0, nullable=False), id="datetime64_zero"
        ),
        param(
            "DateTime64(1)", dt.Timestamp(scale=1, nullable=False), id="datetime64_one"
        ),
    ]
    + [
        param(
            f"DateTime64({scale}, '{tz}')",
            dt.Timestamp(scale=scale, timezone=tz, nullable=False),
            id=f"datetime64_{scale}_{tz}",
        )
        for scale in range(10)
        for tz in ("UTC", "America/New_York", "America/Chicago", "America/Los_Angeles")
    ],
)
def test_parse_type(ch_type, ibis_type):
    parsed_ibis_type = ClickHouseType.from_string(ch_type)
    assert parsed_ibis_type == ibis_type


false = st.just(False)

map_key_types = (
    its.string_dtype(nullable=false)
    | its.integer_dtypes(nullable=false)
    | its.date_dtype(nullable=false)
    | its.timestamp_dtype(scale=st.integers(0, 9), nullable=false)
)

roundtrippable_types = st.deferred(
    lambda: (
        its.null_dtype
        | its.boolean_dtype()
        | its.integer_dtypes()
        | st.just(dt.Float32())
        | st.just(dt.Float64())
        | its.decimal_dtypes()
        | its.string_dtype()
        | its.json_dtype()
        | its.inet_dtype()
        | its.uuid_dtype()
        | its.date_dtype()
        | its.time_dtype()
        | its.timestamp_dtype(scale=st.integers(0, 9))
        | its.array_dtypes(roundtrippable_types)
        | its.map_dtypes(map_key_types, roundtrippable_types, nullable=false)
    )
)


@h.given(roundtrippable_types)
def test_type_roundtrip(ibis_type):
    type_string = ClickHouseType.to_string(ibis_type)
    parsed_ibis_type = ClickHouseType.from_string(type_string)
    assert parsed_ibis_type == ibis_type



from __future__ import annotations

import datetime
import time
from operator import floordiv, methodcaller, truediv

import pandas as pd
import pytest
from pytest import param

import ibis
import ibis.common.exceptions as com
import ibis.expr.datatypes as dt
import ibis.expr.operations as ops
from ibis import _
from ibis.common.annotations import ValidationError

to_sql = ibis.bigquery.compile


@pytest.fixture(scope="module")
def alltypes():
    return ibis.table(
        ibis.schema(
            dict(
                id="int32",
                bool_col="boolean",
                tinyint_col="int8",
                smallint_col="int16",
                int_col="int32",
                bigint_col="int64",
                float_col="float32",
                double_col="float64",
                date_string_col="string",
                string_col="string",
                timestamp_col=dt.Timestamp(timezone="UTC"),
                year="int32",
                month="int32",
            )
        ),
        name="functional_alltypes",
    )


@pytest.mark.parametrize(
    ("case", "dtype"),
    [
        param(datetime.date(2017, 1, 1), dt.date, id="date"),
        param(pd.Timestamp("2017-01-01"), dt.date, id="timestamp_date"),
        param("2017-01-01", dt.date, id="string_date"),
        param(datetime.datetime(2017, 1, 1, 4, 55, 59), dt.timestamp, id="datetime"),
        param("2017-01-01 04:55:59", dt.timestamp, id="string_timestamp"),
        param(pd.Timestamp("2017-01-01 04:55:59"), dt.timestamp, id="timestamp"),
    ],
)
def test_literal_year(case, dtype, snapshot):
    expr = ibis.literal(case, type=dtype).year()
    snapshot.assert_match(to_sql(expr), "out.sql")


@pytest.mark.parametrize(
    ("case", "dtype"),
    [
        param(datetime.date(2017, 1, 1), dt.date, id="date"),
        param(pd.Timestamp("2017-01-01"), dt.date, id="timestamp_date"),
        param("2017-01-01", dt.date, id="string_date"),
        param(datetime.datetime(2017, 1, 1, 4, 55, 59), dt.timestamp, id="datetime"),
        param("2017-01-01 04:55:59", dt.timestamp, id="string_timestamp"),
        param(pd.Timestamp("2017-01-01 04:55:59"), dt.timestamp, id="timestamp"),
    ],
)
def test_day_of_week(case, dtype, snapshot):
    date_var = ibis.literal(case, type=dtype)

    expr = date_var.day_of_week.index()
    snapshot.assert_match(to_sql(expr), "index.sql")

    expr = date_var.day_of_week.full_name()
    snapshot.assert_match(to_sql(expr), "name.sql")


@pytest.mark.parametrize(
    ("case", "dtype"),
    [
        param("test of hash", dt.string, id="string"),
        param(b"test of hash", dt.binary, id="binary"),
    ],
)
def test_hash(case, dtype, snapshot):
    var = ibis.literal(case, type=dtype)
    expr = var.hash()
    snapshot.assert_match(to_sql(expr), "out.sql")


@pytest.mark.parametrize(("case", "dtype"), [("test", "string"), (b"test", "binary")])
@pytest.mark.parametrize("how", ["md5", "sha1", "sha256", "sha512"])
def test_hashbytes(case, how, dtype, snapshot):
    var = ibis.literal(case, type=dtype)
    expr = var.hashbytes(how=how).name("tmp")
    snapshot.assert_match(to_sql(expr), "out.sql")


@pytest.mark.parametrize(
    ("case", "unit"),
    (
        param(123456789, "s", id="s"),
        param(-123456789, "ms", id="ms"),
        param(123456789, "us", id="us"),
        param(1234567891011, "ns", id="ns"),
    ),
)
def test_integer_to_timestamp(case, unit, snapshot):
    expr = ibis.literal(case, type=dt.int64).to_timestamp(unit=unit).name("tmp")
    snapshot.assert_match(to_sql(expr), "out.sql")


@pytest.mark.parametrize(
    ("case",),
    [
        param("a\\b\\c", id="escape_backslash"),
        param("a\ab\bc\fd\ne\rf\tg\vh", id="escape_ascii_sequences"),
        param("a'b\"c", id="escape_quote"),
        param("`~!@#$%^&*()_=+-|[]{};:/?<>", id="not_escape_special_characters"),
    ],
)
def test_literal_string(case, snapshot):
    expr = ibis.literal(case)
    snapshot.assert_match(to_sql(expr), "out.sql")


@pytest.mark.parametrize(
    ("case", "dtype"),
    [
        param(datetime.datetime(2017, 1, 1, 4, 55, 59), dt.timestamp, id="datetime"),
        param("2017-01-01 04:55:59", dt.timestamp, id="string_timestamp"),
        param(pd.Timestamp("2017-01-01 04:55:59"), dt.timestamp, id="timestamp"),
        param(datetime.time(4, 55, 59), dt.time, id="time"),
        param("04:55:59", dt.time, id="string_time"),
    ],
)
def test_literal_timestamp_or_time(case, dtype, snapshot):
    expr = ibis.literal(case, type=dtype).hour().name("tmp")
    snapshot.assert_match(to_sql(expr), "out.sql")


def test_projection_fusion_only_peeks_at_immediate_parent(snapshot):
    schema = [
        ("file_date", "timestamp"),
        ("PARTITIONTIME", "date"),
        ("val", "int64"),
    ]
    table = ibis.table(schema, name="unbound_table")
    table = table[table.PARTITIONTIME < ibis.date("2017-01-01")]
    table = table.mutate(file_date=table.file_date.cast("date"))
    table = table[table.file_date < ibis.date("2017-01-01")]
    table = table.mutate(XYZ=table.val * 2)
    expr = table.join(table.view())[table]
    snapshot.assert_match(to_sql(expr), "out.sql")


unit_full_names = {
    "Y": "year",
    "Q": "quarter",
    "M": "month",
    "W": "week",
    "D": "day",
    "h": "hour",
    "m": "minute",
    "s": "second",
    "ms": "millis",
    "us": "micros",
}


@pytest.mark.parametrize(
    ("unit", "func"),
    [
        ("Y", "timestamp"),
        ("Q", "timestamp"),
        ("M", "timestamp"),
        ("W", "timestamp"),
        ("D", "timestamp"),
        ("h", "timestamp"),
        ("m", "timestamp"),
        ("s", "timestamp"),
        ("ms", "timestamp"),
        ("us", "timestamp"),
        ("Y", "date"),
        ("Q", "date"),
        ("M", "date"),
        ("W", "date"),
        ("D", "date"),
        ("h", "time"),
        ("m", "time"),
        ("s", "time"),
        ("ms", "time"),
        ("us", "time"),
    ],
    ids=lambda p: unit_full_names.get(p, p),
)
def test_temporal_truncate(unit, func, snapshot):
    dtype = getattr(dt, func)
    t = ibis.table([("a", dtype)], name="t")
    expr = t.a.truncate(unit).name("tmp")
    snapshot.assert_match(to_sql(expr), "out.sql")


@pytest.mark.parametrize("kind", ["date", "time"])
def test_extract_temporal_from_timestamp(kind, snapshot):
    t = ibis.table([("ts", dt.timestamp)], name="t")
    expr = getattr(t.ts, kind)().name("tmp")
    snapshot.assert_match(to_sql(expr), "out.sql")


def test_now(snapshot):
    expr = ibis.now()
    snapshot.assert_match(to_sql(expr), "out.sql")


def test_binary(snapshot):
    t = ibis.table([("value", "double")], name="t")
    expr = t["value"].cast(dt.binary)
    snapshot.assert_match(to_sql(expr), "out.sql")


def test_substring(snapshot):
    t = ibis.table([("value", "string")], name="t")
    expr = t["value"].substr(3, 1).name("tmp")
    snapshot.assert_match(to_sql(expr), "out.sql")


def test_substring_neg_length():
    t = ibis.table([("value", "string")], name="t")
    expr = t["value"].substr(3, -1).name("tmp")
    with pytest.raises(
        Exception, match=r"Length parameter must be a non-negative value; got -1"
    ):
        to_sql(expr)


def test_bucket(snapshot):
    t = ibis.table([("value", "double")], name="t")
    buckets = [0, 1, 3]
    expr = t.value.bucket(buckets).name("tmp")
    snapshot.assert_match(to_sql(expr), "out.sql")


@pytest.mark.parametrize(
    ("kind", "begin", "end"),
    [
        param("preceding", None, 1, id="preceding"),
        param("following", 1, None, id="following"),
    ],
)
def test_window_unbounded(kind, begin, end, snapshot):
    t = ibis.table([("a", "int64")], name="t")
    kwargs = {kind: (begin, end)}
    expr = t.a.sum().over(ibis.window(**kwargs)).name("tmp")
    snapshot.assert_match(to_sql(expr), "out.sql")


def test_large_compile():
    """Tests that compiling a large expression tree finishes within a
    reasonable amount of time."""
    num_columns = 20
    num_joins = 7

    class MockBackend(ibis.backends.bigquery.Backend):
        pass

    names = [f"col_{i}" for i in range(num_columns)]
    schema = ibis.Schema(dict.fromkeys(names, "string"))
    ibis_client = MockBackend()
    table = ops.SQLQueryResult("select * from t", schema, ibis_client).to_expr()
    for _ in range(num_joins):  # noqa: F402
        table = table.mutate(dummy=ibis.literal(""))
        table = table.left_join(table, ["dummy"])[[table]]

    start = time.time()
    table.compile()
    delta = time.time() - start
    assert delta < 10


@pytest.mark.parametrize(
    ("operation", "keywords"),
    [
        param("union", {"distinct": False}, id="union_all"),
        param("union", {"distinct": True}, id="union_distinct"),
        param("intersect", {}, id="intersect"),
        param("difference", {}, id="difference"),
    ],
)
def test_set_operation(operation, keywords, snapshot):
    t0 = ibis.table([("a", "int64")], name="t0")
    t1 = ibis.table([("a", "int64")], name="t1")
    expr = getattr(t0, operation)(t1, **keywords)
    snapshot.assert_match(to_sql(expr), "out.sql")


def test_geospatial_point(snapshot):
    t = ibis.table([("lon", "float64"), ("lat", "float64")], name="t")
    expr = t.lon.point(t.lat).name("tmp")
    snapshot.assert_match(to_sql(expr), "out.sql")


def test_geospatial_azimuth(snapshot):
    t = ibis.table([("p0", "point"), ("p1", "point")], name="t")
    expr = t.p0.azimuth(t.p1).name("tmp")
    snapshot.assert_match(to_sql(expr), "out.sql")


def test_geospatial_unary_union(snapshot):
    t = ibis.table([("geog", "geography")], name="t")
    expr = t.geog.unary_union().name("tmp")
    snapshot.assert_match(to_sql(expr), "out.sql")


@pytest.mark.parametrize(
    ("operation", "keywords"),
    [
        param("area", {}, id="aread"),
        param("as_binary", {}, id="as_binary"),
        param("as_text", {}, id="as_text"),
        param("buffer", {"radius": 5.2}, id="buffer"),
        param("centroid", {}, id="centroid"),
        param("end_point", {}, id="end_point"),
        param("geometry_type", {}, id="geometry_type"),
        param("length", {}, id="length"),
        param("n_points", {}, id="npoints"),
        param("perimeter", {}, id="perimeter"),
        param("point_n", {"n": 3}, id="point_n"),
        param("start_point", {}, id="start_point"),
    ],
)
def test_geospatial_unary(operation, keywords, snapshot):
    t = ibis.table([("geog", "geography")], name="t")
    expr = getattr(t.geog, operation)(**keywords).name("tmp")
    snapshot.assert_match(to_sql(expr), "out.sql")


@pytest.mark.parametrize(
    ("operation", "keywords"),
    [
        param("contains", {}, id="contains"),
        param("covers", {}, id="covers"),
        param("covered_by", {}, id="covered_by"),
        param("d_within", {"distance": 5.2}, id="d_within"),
        param("difference", {}, id="difference"),
        param("disjoint", {}, id="disjoint"),
        param("distance", {}, id="distance"),
        param("geo_equals", {}, id="geo_equals"),
        param("intersection", {}, id="intersection"),
        param("intersects", {}, id="intersects"),
        param("max_distance", {}, id="max_distance"),
        param("touches", {}, id="touches"),
        param("union", {}, id="union"),
        param("within", {}, id="within"),
    ],
)
def test_geospatial_binary(operation, keywords, snapshot):
    t = ibis.table([("geog0", "geography"), ("geog1", "geography")], name="t")
    expr = getattr(t.geog0, operation)(t.geog1, **keywords).name("tmp")
    snapshot.assert_match(to_sql(expr), "out.sql")


@pytest.mark.parametrize("operation", ["x_max", "x_min", "y_max", "y_min"])
def test_geospatial_minmax(operation, snapshot):
    t = ibis.table([("geog", "geography")], name="t")
    expr = getattr(t.geog, operation)().name("tmp")
    snapshot.assert_match(to_sql(expr), "out.sql")


@pytest.mark.parametrize("dimension_name", ["x", "y"])
def test_geospatial_xy(dimension_name, snapshot):
    t = ibis.table([("pt", "point")], name="t")
    expr = getattr(t.pt, dimension_name)().name("tmp")
    snapshot.assert_match(to_sql(expr), "out.sql")


def test_geospatial_simplify(snapshot):
    t = ibis.table([("geog", "geography")], name="t")
    expr = t.geog.simplify(5.2, preserve_collapsed=False).name("tmp")
    snapshot.assert_match(to_sql(expr), "out.sql")


def test_geospatial_simplify_error():
    t = ibis.table([("geog", "geography")], name="t")
    expr = t.geog.simplify(5.2, preserve_collapsed=True).name("tmp")
    with pytest.raises(
        Exception, match="simplify does not support preserving collapsed geometries"
    ):
        to_sql(expr)


def test_timestamp_accepts_date_literals(alltypes):
    date_string = "2009-03-01"
    p = ibis.param(dt.timestamp).name("param_0")
    expr = alltypes.mutate(param=p)
    params = {p: date_string}
    result = to_sql(expr, params=params)
    assert "2009-03-01T00:00:00" in result


@pytest.mark.parametrize("distinct", [True, False])
def test_union(alltypes, distinct, snapshot):
    expr = alltypes.union(alltypes, distinct=distinct)
    snapshot.assert_match(to_sql(expr), "out.sql")


@pytest.mark.parametrize("op", [truediv, floordiv])
def test_divide_by_zero(alltypes, op, snapshot):
    expr = op(alltypes.double_col, 0)
    snapshot.assert_match(to_sql(expr), "out.sql")


def test_identical_to(alltypes, snapshot):
    expr = alltypes[
        _.string_col.identical_to("a") & _.date_string_col.identical_to("b")
    ]
    snapshot.assert_match(to_sql(expr), "out.sql")


def test_to_timestamp_no_timezone(alltypes, snapshot):
    expr = alltypes.date_string_col.to_timestamp("%F")
    snapshot.assert_match(to_sql(expr), "out.sql")


def test_to_timestamp_timezone(alltypes, snapshot):
    expr = (alltypes.date_string_col + " America/New_York").to_timestamp("%F %Z")
    snapshot.assert_match(to_sql(expr), "out.sql")


@pytest.mark.parametrize(
    "window",
    [
        param(
            ibis.window(
                preceding=1, following=0, group_by="year", order_by="timestamp_col"
            ),
            id="prec_current",
        ),
        param(
            ibis.window(
                preceding=0, following=2, group_by="year", order_by="timestamp_col"
            ),
            id="current_foll",
        ),
        param(
            ibis.window(preceding=(4, 2), group_by="year", order_by="timestamp_col"),
            id="prec_prec",
        ),
    ],
)
def test_window_function(alltypes, window, snapshot):
    expr = alltypes.mutate(win_avg=_.float_col.mean().over(window))
    snapshot.assert_match(to_sql(expr), "out.sql")


@pytest.mark.parametrize(
    "window",
    [
        param(
            ibis.range_window(
                preceding=1, following=0, group_by="year", order_by="month"
            ),
            id="prec_foll",
        ),
        param(
            ibis.range_window(
                preceding=(4, 2), group_by="year", order_by="timestamp_col"
            ),
            id="prec_prec",
        ),
    ],
)
def test_range_window_function(alltypes, window, snapshot):
    expr = alltypes.mutate(two_month_avg=_.float_col.mean().over(window))
    snapshot.assert_match(to_sql(expr), "out.sql")


@pytest.mark.parametrize(
    "preceding",
    [
        param(5, id="five"),
        param(
            ibis.interval(nanoseconds=1),
            id="nanos",
            marks=pytest.mark.xfail(raises=com.UnsupportedOperationError),
        ),
        param(ibis.interval(microseconds=1), id="micros"),
        param(ibis.interval(seconds=1), id="seconds"),
        param(ibis.interval(minutes=1), id="minutes"),
        param(ibis.interval(hours=1), id="hours"),
        param(ibis.interval(days=1), id="days"),
        param(2 * ibis.interval(days=1), id="two_days"),
        param(ibis.interval(weeks=1), id="week"),
    ],
)
def test_trailing_range_window(alltypes, preceding, snapshot):
    w = ibis.trailing_range_window(preceding=preceding, order_by="timestamp_col")
    expr = alltypes.mutate(win_avg=_.float_col.mean().over(w))
    snapshot.assert_match(to_sql(expr), "out.sql")


@pytest.mark.parametrize("distinct1", [True, False])
@pytest.mark.parametrize("distinct2", [True, False])
def test_union_cte(alltypes, distinct1, distinct2, snapshot):
    t = alltypes
    expr1 = t.group_by(t.string_col).agg(metric=t.double_col.sum())
    expr2 = expr1.view()
    expr3 = expr1.view()
    expr = expr1.union(expr2, distinct=distinct1).union(expr3, distinct=distinct2)
    snapshot.assert_match(to_sql(expr), "out.sql")


@pytest.mark.parametrize("funcname", ["sum", "mean"])
def test_bool_reducers(alltypes, funcname, snapshot):
    method = methodcaller(funcname)
    expr = method(alltypes.bool_col)
    snapshot.assert_match(to_sql(expr), "out.sql")


def test_bool_reducers_where_simple(alltypes, snapshot):
    b = alltypes.bool_col
    m = alltypes.month
    expr = b.mean(where=m > 6)
    snapshot.assert_match(to_sql(expr), "out.sql")


def test_bool_reducers_where_conj(alltypes, snapshot):
    b = alltypes.bool_col
    m = alltypes.month
    expr2 = b.sum(where=((m > 6) & (m < 10)))
    snapshot.assert_match(to_sql(expr2), "out.sql")


@pytest.mark.parametrize("agg", ["approx_median", "approx_nunique"])
@pytest.mark.parametrize(
    "where",
    [param(lambda _: None, id="no_filter"), param(lambda t: t.month > 0, id="filter")],
)
def test_approx(alltypes, agg, where, snapshot):
    d = alltypes.double_col
    method = methodcaller(agg, where=where(alltypes))
    expr = method(d)
    snapshot.assert_match(to_sql(expr), "out.sql")


@pytest.mark.parametrize("funcname", ["bit_and", "bit_or", "bit_xor"])
@pytest.mark.parametrize(
    "where",
    [
        param(lambda _: None, id="no_filter"),
        param(lambda t: t.bigint_col > 0, id="filter"),
    ],
)
def test_bit(alltypes, funcname, where, snapshot):
    method = methodcaller(funcname, where=where(alltypes))
    expr = method(alltypes.int_col)
    snapshot.assert_match(to_sql(expr), "out.sql")


@pytest.mark.parametrize("how", ["pop", "sample"])
def test_cov(alltypes, how, snapshot):
    d = alltypes.double_col
    expr = d.cov(d, how=how)
    snapshot.assert_match(to_sql(expr), "out.sql")


def test_cov_invalid_how(alltypes):
    d = alltypes.double_col

    with pytest.raises(ValidationError):
        d.cov(d, how="error")


def test_compile_toplevel(snapshot):
    t = ibis.table([("foo", "double")], name="t0")

    # it works!
    expr = t.foo.sum()
    result = to_sql(expr)
    snapshot.assert_match(result, "out.sql")


def test_scalar_param_scope(alltypes):
    t = alltypes
    param = ibis.param("timestamp")
    result = to_sql(t.mutate(param=param), params={param: "2017-01-01"})
    assert "2017-01-01T00:00:00" in result


def test_cast_float_to_int(alltypes, snapshot):
    expr = alltypes.double_col.cast("int64")
    result = to_sql(expr)
    snapshot.assert_match(result, "out.sql")


def test_unnest(snapshot):
    table = ibis.table(
        dict(
            rowindex="int",
            repeated_struct_col=dt.Array(
                dt.Struct(
                    dict(
                        nested_struct_col=dt.Array(
                            dt.Struct(
                                dict(
                                    doubly_nested_array="array<int>",
                                    doubly_nested_field="string",
                                )
                            )
                        )
                    )
                )
            ),
        ),
        name="array_test",
    )
    repeated_struct_col = table.repeated_struct_col

    # Works as expected :-)
    result = ibis.bigquery.compile(
        table.select("rowindex", repeated_struct_col.unnest())
    )
    snapshot.assert_match(result, "out_one_unnest.sql")

    result = ibis.bigquery.compile(
        table.select(
            "rowindex", level_one=repeated_struct_col.unnest().nested_struct_col
        ).select(level_two=lambda t: t.level_one.unnest())
    )
    snapshot.assert_match(result, "out_two_unnests.sql")



"""Risingwave backend."""

from __future__ import annotations

from functools import partial
from itertools import repeat
from typing import TYPE_CHECKING

import psycopg2
import sqlglot as sg
import sqlglot.expressions as sge
from psycopg2 import extras

import ibis
import ibis.common.exceptions as com
import ibis.expr.operations as ops
import ibis.expr.types as ir
from ibis import util
from ibis.backends.postgres import Backend as PostgresBackend
from ibis.backends.risingwave.compiler import RisingwaveCompiler

if TYPE_CHECKING:
    import pandas as pd
    import pyarrow as pa


class Backend(PostgresBackend):
    name = "risingwave"
    compiler = RisingwaveCompiler()
    supports_python_udfs = False

    def do_connect(
        self,
        host: str | None = None,
        user: str | None = None,
        password: str | None = None,
        port: int = 5432,
        database: str | None = None,
        schema: str | None = None,
    ) -> None:
        """Create an Ibis client connected to RisingWave database.

        Parameters
        ----------
        host
            Hostname
        user
            Username
        password
            Password
        port
            Port number
        database
            Database to connect to
        schema
            RisingWave schema to use. If `None`, use the default `search_path`.

        Examples
        --------
        >>> import os
        >>> import getpass
        >>> import ibis
        >>> host = os.environ.get("IBIS_TEST_RISINGWAVE_HOST", "localhost")
        >>> user = os.environ.get("IBIS_TEST_RISINGWAVE_USER", getpass.getuser())
        >>> password = os.environ.get("IBIS_TEST_RISINGWAVE_PASSWORD")
        >>> database = os.environ.get("IBIS_TEST_RISINGWAVE_DATABASE", "dev")
        >>> con = connect(database=database, host=host, user=user, password=password)
        >>> con.list_tables()  # doctest: +ELLIPSIS
        [...]
        >>> t = con.table("functional_alltypes")
        >>> t
        RisingwaveTable[table]
          name: functional_alltypes
          schema:
            id : int32
            bool_col : boolean
            tinyint_col : int16
            smallint_col : int16
            int_col : int32
            bigint_col : int64
            float_col : float32
            double_col : float64
            date_string_col : string
            string_col : string
            timestamp_col : timestamp
            year : int32
            month : int32

        """

        self.con = psycopg2.connect(
            host=host,
            port=port,
            user=user,
            password=password,
            database=database,
            options=(f"-csearch_path={schema}" * (schema is not None)) or None,
        )

        with self.begin() as cur:
            cur.execute("SET TIMEZONE = UTC")

    def create_table(
        self,
        name: str,
        obj: pd.DataFrame | pa.Table | ir.Table | None = None,
        *,
        schema: ibis.Schema | None = None,
        database: str | None = None,
        temp: bool = False,
        overwrite: bool = False,
    ):
        """Create a table in Risingwave.

        Parameters
        ----------
        name
            Name of the table to create
        obj
            The data with which to populate the table; optional, but at least
            one of `obj` or `schema` must be specified
        schema
            The schema of the table to create; optional, but at least one of
            `obj` or `schema` must be specified
        database
            The name of the database in which to create the table; if not
            passed, the current database is used.
        temp
            Create a temporary table
        overwrite
            If `True`, replace the table if it already exists, otherwise fail
            if the table exists

        """
        if obj is None and schema is None:
            raise ValueError("Either `obj` or `schema` must be specified")

        if database is not None and database != self.current_database:
            raise com.UnsupportedOperationError(
                f"Creating tables in other databases is not supported by {self.name}"
            )
        else:
            database = None

        properties = []

        if temp:
            properties.append(sge.TemporaryProperty())

        if obj is not None:
            if not isinstance(obj, ir.Expr):
                table = ibis.memtable(obj)
            else:
                table = obj

            self._run_pre_execute_hooks(table)

            query = self._to_sqlglot(table)
        else:
            query = None

        column_defs = [
            sge.ColumnDef(
                this=sg.to_identifier(colname, quoted=self.compiler.quoted),
                kind=self.compiler.type_mapper.from_ibis(typ),
                constraints=(
                    None
                    if typ.nullable
                    else [sge.ColumnConstraint(kind=sge.NotNullColumnConstraint())]
                ),
            )
            for colname, typ in (schema or table.schema()).items()
        ]

        if overwrite:
            temp_name = util.gen_name(f"{self.name}_table")
        else:
            temp_name = name

        table = sg.table(temp_name, catalog=database, quoted=self.compiler.quoted)
        target = sge.Schema(this=table, expressions=column_defs)

        create_stmt = sge.Create(
            kind="TABLE",
            this=target,
            properties=sge.Properties(expressions=properties),
        )

        this = sg.table(name, catalog=database, quoted=self.compiler.quoted)
        with self._safe_raw_sql(create_stmt) as cur:
            if query is not None:
                insert_stmt = sge.Insert(this=table, expression=query).sql(self.dialect)
                cur.execute(insert_stmt)

            if overwrite:
                cur.execute(
                    sge.Drop(kind="TABLE", this=this, exists=True).sql(self.dialect)
                )
                cur.execute(
                    f"ALTER TABLE {table.sql(self.dialect)} RENAME TO {this.sql(self.dialect)}"
                )

        if schema is None:
            return self.table(name, schema=database)

        # preserve the input schema if it was provided
        return ops.DatabaseTable(
            name, schema=schema, source=self, namespace=ops.Namespace(database=database)
        ).to_expr()

    def _register_in_memory_table(self, op: ops.InMemoryTable) -> None:
        schema = op.schema
        if null_columns := [col for col, dtype in schema.items() if dtype.is_null()]:
            raise com.IbisTypeError(
                f"{self.name} cannot yet reliably handle `null` typed columns; "
                f"got null typed columns: {null_columns}"
            )

        # only register if we haven't already done so
        if (name := op.name) not in self.list_tables():
            quoted = self.compiler.quoted
            column_defs = [
                sg.exp.ColumnDef(
                    this=sg.to_identifier(colname, quoted=quoted),
                    kind=self.compiler.type_mapper.from_ibis(typ),
                    constraints=(
                        None
                        if typ.nullable
                        else [
                            sg.exp.ColumnConstraint(
                                kind=sg.exp.NotNullColumnConstraint()
                            )
                        ]
                    ),
                )
                for colname, typ in schema.items()
            ]

            create_stmt = sg.exp.Create(
                kind="TABLE",
                this=sg.exp.Schema(
                    this=sg.to_identifier(name, quoted=quoted), expressions=column_defs
                ),
            )
            create_stmt_sql = create_stmt.sql(self.dialect)

            columns = schema.keys()
            df = op.data.to_frame()
            data = df.itertuples(index=False)
            cols = ", ".join(
                ident.sql(self.dialect)
                for ident in map(partial(sg.to_identifier, quoted=quoted), columns)
            )
            specs = ", ".join(repeat("%s", len(columns)))
            table = sg.table(name, quoted=quoted)
            sql = f"INSERT INTO {table.sql(self.dialect)} ({cols}) VALUES ({specs})"
            with self.begin() as cur:
                cur.execute(create_stmt_sql)
                extras.execute_batch(cur, sql, data, 128)



from __future__ import annotations

import pyspark
import pyspark.sql.types as pt
from packaging.version import parse as vparse

import ibis.common.exceptions as com
import ibis.expr.datatypes as dt
import ibis.expr.schema as sch
from ibis.formats import SchemaMapper, TypeMapper

# DayTimeIntervalType introduced in Spark 3.2 (at least) but didn't show up in
# PySpark until version 3.3
PYSPARK_33 = vparse(pyspark.__version__) >= vparse("3.3")
PYSPARK_35 = vparse(pyspark.__version__) >= vparse("3.5")


_from_pyspark_dtypes = {
    pt.BinaryType: dt.Binary,
    pt.BooleanType: dt.Boolean,
    pt.ByteType: dt.Int8,
    pt.DateType: dt.Date,
    pt.DoubleType: dt.Float64,
    pt.FloatType: dt.Float32,
    pt.IntegerType: dt.Int32,
    pt.LongType: dt.Int64,
    pt.NullType: dt.Null,
    pt.ShortType: dt.Int16,
    pt.StringType: dt.String,
    pt.TimestampType: dt.Timestamp,
}

_to_pyspark_dtypes = {v: k for k, v in _from_pyspark_dtypes.items()}
_to_pyspark_dtypes[dt.JSON] = pt.StringType
_to_pyspark_dtypes[dt.UUID] = pt.StringType


if PYSPARK_33:
    _pyspark_interval_units = {
        pt.DayTimeIntervalType.SECOND: "s",
        pt.DayTimeIntervalType.MINUTE: "m",
        pt.DayTimeIntervalType.HOUR: "h",
        pt.DayTimeIntervalType.DAY: "D",
    }


class PySparkType(TypeMapper):
    @classmethod
    def to_ibis(cls, typ, nullable=True):
        """Convert a pyspark type to an ibis type."""
        if isinstance(typ, pt.DecimalType):
            return dt.Decimal(typ.precision, typ.scale, nullable=nullable)
        elif isinstance(typ, pt.ArrayType):
            return dt.Array(cls.to_ibis(typ.elementType), nullable=nullable)
        elif isinstance(typ, pt.MapType):
            return dt.Map(
                cls.to_ibis(typ.keyType),
                cls.to_ibis(typ.valueType),
                nullable=nullable,
            )
        elif isinstance(typ, pt.StructType):
            fields = {f.name: cls.to_ibis(f.dataType) for f in typ.fields}

            return dt.Struct(fields, nullable=nullable)
        elif PYSPARK_33 and isinstance(typ, pt.DayTimeIntervalType):
            if (
                typ.startField == typ.endField
                and typ.startField in _pyspark_interval_units
            ):
                unit = _pyspark_interval_units[typ.startField]
                return dt.Interval(unit, nullable=nullable)
            else:
                raise com.IbisTypeError(f"{typ!r} couldn't be converted to Interval")
        elif PYSPARK_35 and isinstance(typ, pt.TimestampNTZType):
            return dt.Timestamp(nullable=nullable)
        elif isinstance(typ, pt.UserDefinedType):
            return cls.to_ibis(typ.sqlType(), nullable=nullable)
        else:
            try:
                return _from_pyspark_dtypes[type(typ)](nullable=nullable)
            except KeyError:
                raise NotImplementedError(
                    f"Unable to convert type {typ} of type {type(typ)} to an ibis type."
                )

    @classmethod
    def from_ibis(cls, dtype):
        if dtype.is_decimal():
            return pt.DecimalType(dtype.precision, dtype.scale)
        elif dtype.is_array():
            element_type = cls.from_ibis(dtype.value_type)
            contains_null = dtype.value_type.nullable
            return pt.ArrayType(element_type, contains_null)
        elif dtype.is_map():
            key_type = cls.from_ibis(dtype.key_type)
            value_type = cls.from_ibis(dtype.value_type)
            value_contains_null = dtype.value_type.nullable
            return pt.MapType(key_type, value_type, value_contains_null)
        elif dtype.is_struct():
            fields = [
                pt.StructField(n, cls.from_ibis(t), t.nullable)
                for n, t in dtype.fields.items()
            ]
            return pt.StructType(fields)
        else:
            try:
                return _to_pyspark_dtypes[type(dtype)]()
            except KeyError:
                raise com.IbisTypeError(
                    f"Unable to convert dtype {dtype!r} to pyspark type"
                )


class PySparkSchema(SchemaMapper):
    @classmethod
    def from_ibis(cls, schema):
        fields = [
            pt.StructField(name, PySparkType.from_ibis(dtype), dtype.nullable)
            for name, dtype in schema.items()
        ]
        return pt.StructType(fields)

    @classmethod
    def to_ibis(cls, schema):
        return sch.Schema({name: PySparkType.to_ibis(typ) for name, typ in schema})



from __future__ import annotations

import hypothesis as h
import pytest

import ibis.expr.datatypes as dt
import ibis.tests.strategies as ibst


@h.given(ibst.signed_integer_dtypes(), ibst.signed_integer_dtypes())
def test_signed_integer_castable_to_signed_integer(from_, to):
    if from_.nbytes > to.nbytes:
        assert not from_.castable(to)
    else:
        assert from_.castable(to)


@h.given(ibst.unsigned_integer_dtypes(), ibst.unsigned_integer_dtypes())
def test_unsigned_integer_castable_to_unsigned_integer(from_, to):
    if from_.nbytes > to.nbytes:
        assert not from_.castable(to)
    else:
        assert from_.castable(to)


# TODO(kszucs): unsigned to signed implicit cast is currently allowed to not
# break the integral promotion rule logic in rules.py
# @h.given(ibst.signed_integer_dtypes(), ibst.unsigned_integer_dtypes())
# def test_signed_integer_not_castable_to_unsigned_integer(from_, to):
#     assert not from_.castable(to)


@h.given(ibst.integer_dtypes(), ibst.floating_dtypes())
def test_integer_castable_to_floating(from_, to):
    assert from_.castable(to)


# TODO(kszucs): we could be more pedantic here considering the precision
@h.given(ibst.integer_dtypes(), ibst.decimal_dtypes())
def test_integer_castable_to_decimal(from_, to):
    assert from_.castable(to)


@h.given(ibst.integer_dtypes(), ibst.boolean_dtype())
def test_integer_castable_to_boolean(from_, to):
    assert from_.castable(to, value=0)
    assert from_.castable(to, value=1)
    assert not from_.castable(to, value=-1)
    assert not from_.castable(to)


@pytest.mark.parametrize(
    ("source", "target"),
    [
        (dt.string, dt.uuid),
        (dt.uuid, dt.string),
        (dt.null, dt.date),
        (dt.int8, dt.int64),
        (dt.int8, dt.Decimal(12, 2)),
        # (dt.int16, dt.uint64),
        (dt.int32, dt.int32),
        (dt.int32, dt.int64),
        (dt.uint32, dt.uint64),
        (dt.uint32, dt.int64),
        (dt.uint32, dt.Decimal(12, 2)),
        (dt.uint32, dt.float32),
        (dt.uint32, dt.float64),
        (dt.Interval("s"), dt.Interval("s")),
    ],
)
def test_implicitly_castable_primitives(source, target):
    assert source.castable(target)


@pytest.mark.parametrize(
    ("source", "target"),
    [
        (dt.string, dt.null),
        (dt.int32, dt.int16),
        (dt.int32, dt.uint16),
        (dt.uint64, dt.int16),
        (dt.uint64, dt.uint16),
        # (dt.uint64, dt.int64), TODO: https://github.com/ibis-project/ibis/issues/7331
        (dt.Decimal(12, 2), dt.int32),
        (dt.timestamp, dt.boolean),
        (dt.Interval("s"), dt.Interval("ns")),
    ],
)
def test_implicitly_uncastable_primitives(source, target):
    assert not source.castable(target)


@pytest.mark.parametrize("value", [0, 1])
def test_implicitly_castable_int_to_bool(value):
    assert dt.int8.castable(dt.boolean, value=value)


@pytest.mark.parametrize(
    ("source", "target", "value"),
    [(dt.int8, dt.boolean, 3), (dt.int8, dt.boolean, -1)],
)
def test_implicitly_uncastable_values(source, target, value):
    assert not source.castable(target, value=value)



from __future__ import annotations

from unittest import mock

import google.api_core.client_options
import google.api_core.exceptions as gexc
import pydata_google_auth
import pytest
from google.auth import credentials as auth
from google.cloud import bigquery as bq
from google.cloud import bigquery_storage_v1 as bqstorage

import ibis


def test_repeated_project_name(project_id, credentials):
    con = ibis.bigquery.connect(
        project_id=project_id,
        dataset_id=f"{project_id}.testing",
        credentials=credentials,
    )
    try:
        assert "functional_alltypes" in con.list_tables()
    except gexc.Forbidden:
        pytest.skip("Cannot access BigQuery")


def test_project_id_different_from_default_credentials(monkeypatch):
    creds = mock.create_autospec(auth.Credentials)

    def mock_credentials(*args, **kwargs):
        return creds, "default-project-id"

    monkeypatch.setattr(pydata_google_auth, "default", mock_credentials)
    con = ibis.bigquery.connect(
        project_id="explicit-project-id",
    )
    assert con.billing_project == "explicit-project-id"


def test_without_dataset(project_id, credentials):
    con = ibis.bigquery.connect(project_id=project_id, credentials=credentials)
    with pytest.raises(ValueError, match="Unable to determine BigQuery"):
        con.list_tables()


def test_application_name_sets_user_agent(project_id, credentials, monkeypatch):
    mock_client = mock.create_autospec(bq.Client)
    monkeypatch.setattr(bq, "Client", mock_client)
    ibis.bigquery.connect(
        project_id=project_id,
        dataset_id="bigquery-public-data.stackoverflow",
        application_name="my-great-app/0.7.0",
        credentials=credentials,
    )
    info = mock_client.call_args[1]["client_info"]
    user_agent = info.to_user_agent()
    assert f" ibis/{ibis.__version__}" in user_agent
    assert "my-great-app/0.7.0 " in user_agent


def test_auth_default(project_id, credentials, monkeypatch):
    mock_calls = []

    def mock_default(*args, **kwargs):
        mock_calls.append((args, kwargs))
        return credentials, project_id

    monkeypatch.setattr(pydata_google_auth, "default", mock_default)

    ibis.bigquery.connect(
        project_id=project_id,
        dataset_id="bigquery-public-data.stackoverflow",
    )

    assert len(mock_calls) == 1
    args, kwargs = mock_calls[0]
    assert len(args) == 1
    scopes = args[0]
    assert scopes == ibis.backends.bigquery.SCOPES
    auth_local_webserver = kwargs["use_local_webserver"]
    auth_cache = kwargs["credentials_cache"]
    assert auth_local_webserver
    assert isinstance(
        auth_cache,
        pydata_google_auth.cache.ReadWriteCredentialsCache,
    )


def test_auth_local_webserver(project_id, credentials, monkeypatch):
    mock_calls = []

    def mock_default(*args, **kwargs):
        mock_calls.append((args, kwargs))
        return credentials, project_id

    monkeypatch.setattr(pydata_google_auth, "default", mock_default)

    ibis.bigquery.connect(
        project_id=project_id,
        dataset_id="bigquery-public-data.stackoverflow",
        auth_local_webserver=False,
    )

    assert len(mock_calls) == 1
    _, kwargs = mock_calls[0]
    auth_local_webserver = kwargs["use_local_webserver"]
    assert not auth_local_webserver


def test_auth_external_data(project_id, credentials, monkeypatch):
    mock_calls = []

    def mock_default(*args, **kwargs):
        mock_calls.append((args, kwargs))
        return credentials, project_id

    monkeypatch.setattr(pydata_google_auth, "default", mock_default)

    ibis.bigquery.connect(
        project_id=project_id,
        dataset_id="bigquery-public-data.stackoverflow",
        auth_external_data=True,
    )

    assert len(mock_calls) == 1
    args, _ = mock_calls[0]
    assert len(args) == 1
    scopes = args[0]
    assert scopes == ibis.backends.bigquery.EXTERNAL_DATA_SCOPES


def test_auth_cache_reauth(project_id, credentials, monkeypatch):
    mock_calls = []

    def mock_default(*args, **kwargs):
        mock_calls.append((args, kwargs))
        return credentials, project_id

    monkeypatch.setattr(pydata_google_auth, "default", mock_default)

    ibis.bigquery.connect(
        project_id=project_id,
        dataset_id="bigquery-public-data.stackoverflow",
        auth_cache="reauth",
    )

    assert len(mock_calls) == 1
    _, kwargs = mock_calls[0]
    auth_cache = kwargs["credentials_cache"]
    assert isinstance(
        auth_cache,
        pydata_google_auth.cache.WriteOnlyCredentialsCache,
    )


def test_auth_cache_none(project_id, credentials, monkeypatch):
    mock_calls = []

    def mock_default(*args, **kwargs):
        mock_calls.append((args, kwargs))
        return credentials, project_id

    monkeypatch.setattr(pydata_google_auth, "default", mock_default)

    ibis.bigquery.connect(
        project_id=project_id,
        dataset_id="bigquery-public-data.stackoverflow",
        auth_cache="none",
    )

    assert len(mock_calls) == 1
    _, kwargs = mock_calls[0]
    auth_cache = kwargs["credentials_cache"]
    assert auth_cache is pydata_google_auth.cache.NOOP


def test_auth_cache_unknown(project_id):
    with pytest.raises(ValueError, match="unexpected value for auth_cache"):
        ibis.bigquery.connect(
            project_id=project_id,
            dataset_id="bigquery-public-data.stackoverflow",
            auth_cache="not_a_real_cache",
        )


def test_client_with_regional_endpoints(
    project_id, credentials, dataset_id, dataset_id_tokyo, region_tokyo
):
    bq_options = google.api_core.client_options.ClientOptions(
        api_endpoint=f"https://{region_tokyo}-bigquery.googleapis.com"
    )
    bq_client = bq.Client(
        client_options=bq_options, project=project_id, credentials=credentials
    )

    # Note there is no protocol specifier for gRPC APIs.
    bqstorage_options = google.api_core.client_options.ClientOptions(
        api_endpoint=f"{region_tokyo}-bigquerystorage.googleapis.com"
    )
    bqstorage_client = bqstorage.BigQueryReadClient(
        client_options=bqstorage_options, credentials=credentials
    )

    con = ibis.bigquery.connect(
        client=bq_client, storage_client=bqstorage_client, project_id=project_id
    )

    # Fails because dataset not in Tokyo.
    with pytest.raises(gexc.NotFound, match=dataset_id):
        con.table(f"{dataset_id}.functional_alltypes")

    # Succeeds because dataset is in Tokyo.
    alltypes = con.table(f"{dataset_id_tokyo}.functional_alltypes")
    df = alltypes.limit(2).execute()
    assert len(df.index) == 2
